# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MobileNetV3 models for Deep Labeling.

Reference:
  Howard, A., Sandler, M., et al. Searching for mobilenetv3. In ICCV, 2019
"""
from typing import Any, Callable, Mapping, Optional, Sequence

import tensorflow as tf
from deeplab2.model import utils
from deeplab2.model.layers import blocks
from deeplab2.model.layers import convolutions

# The default input image channels.
_INPUT_CHANNELS = 3

MNV3Small_BLOCK_SPECS = {
    'spec_name': 'MobileNetV3Small',
    'block_spec_schema': ['block_fn', 'kernel_size', 'strides', 'filters',
                          'activation', 'se_ratio', 'expand_ratio',
                          'is_endpoint'],
    'block_specs': [
        ('conv_bn', 3, 2, 16,
         'hard_swish', None, None, True),
        ('inverted_bottleneck', 3, 2, 16,
         'relu', 0.25, 1, True),
        ('inverted_bottleneck', 3, 2, 24,
         'relu', None, 72. / 16, False),
        ('inverted_bottleneck', 3, 1, 24,
         'relu', None, 88. / 24, True),
        ('inverted_bottleneck', 5, 2, 40,
         'hard_swish', 0.25, 4., False),
        ('inverted_bottleneck', 5, 1, 40,
         'hard_swish', 0.25, 6., False),
        ('inverted_bottleneck', 5, 1, 40,
         'hard_swish', 0.25, 6., False),
        ('inverted_bottleneck', 5, 1, 48,
         'hard_swish', 0.25, 3., False),
        ('inverted_bottleneck', 5, 1, 48,
         'hard_swish', 0.25, 3., True),
        ('inverted_bottleneck', 5, 2, 96,
         'hard_swish', 0.25, 6., False),
        ('inverted_bottleneck', 5, 1, 96,
         'hard_swish', 0.25, 6., False),
        ('inverted_bottleneck', 5, 1, 96,
         'hard_swish', 0.25, 6., False),
        ('conv_bn', 1, 1, 576,
         'hard_swish', None, None, True),
    ]
}

MNV3Large_BLOCK_SPECS = {
    'spec_name': 'MobileNetV3Large',
    'block_spec_schema': ['block_fn', 'kernel_size', 'strides', 'filters',
                          'activation', 'se_ratio', 'expand_ratio',
                          'is_endpoint'],
    'block_specs': [
        ('conv_bn', 3, 2, 16,
         'hard_swish', None, None, False),
        ('inverted_bottleneck', 3, 1, 16,
         'relu', None, 1., True),
        ('inverted_bottleneck', 3, 2, 24,
         'relu', None, 4., False),
        ('inverted_bottleneck', 3, 1, 24,
         'relu', None, 3., True),
        ('inverted_bottleneck', 5, 2, 40,
         'relu', 0.25, 3., False),
        ('inverted_bottleneck', 5, 1, 40,
         'relu', 0.25, 3., False),
        ('inverted_bottleneck', 5, 1, 40,
         'relu', 0.25, 3., True),
        ('inverted_bottleneck', 3, 2, 80,
         'hard_swish', None, 6., False),
        ('inverted_bottleneck', 3, 1, 80,
         'hard_swish', None, 2.5, False),
        ('inverted_bottleneck', 3, 1, 80,
         'hard_swish', None, 2.3, False),
        ('inverted_bottleneck', 3, 1, 80,
         'hard_swish', None, 2.3, False),
        ('inverted_bottleneck', 3, 1, 112,
         'hard_swish', 0.25, 6., False),
        ('inverted_bottleneck', 3, 1, 112,
         'hard_swish', 0.25, 6., True),
        ('inverted_bottleneck', 5, 2, 160,
         'hard_swish', 0.25, 6., False),
        ('inverted_bottleneck', 5, 1, 160,
         'hard_swish', 0.25, 6., False),
        ('inverted_bottleneck', 5, 1, 160,
         'hard_swish', 0.25, 6., False),
        ('conv_bn', 1, 1, 960,
         'hard_swish', None, None, True),
    ]
}

SUPPORTED_SPECS_MAP = {
    'MobileNetV3Large': MNV3Large_BLOCK_SPECS,
    'MobileNetV3Small': MNV3Small_BLOCK_SPECS,
}


# pylint: disable=invalid-name
def _block_spec_decoder(
        specs: Mapping[Any, Any],
        width_multiplier: float,
        divisible_by: int = 8,
        reduce_last_block_filters: bool = False) -> Sequence[Mapping[str, Any]]:
    """Decodes specs for a block.

    Args:
      specs: A `dict` specification of block specs of a mobilenet version.
      width_multiplier: A `float` multiplier for the filter size for all
        convolution ops. The value must be greater than zero. Typical usage will
        be to set this value in (0, 1) to reduce the number of parameters or
        computation cost of the model.
      divisible_by: An `int` that ensures all inner dimensions are divisible by
        this number.
      reduce_last_block_filters: A bool indicates whether to reduce the final
        block's filters.

    Returns:
      A list of block spec in dictionary that defines structure of the layers.
    """

    spec_name = specs['spec_name']
    block_spec_schema = specs['block_spec_schema']
    block_specs = specs['block_specs']

    if not block_specs:
        raise ValueError(
            'The block spec cannot be empty for {} !'.format(spec_name))

    if len(block_specs[0]) != len(block_spec_schema):
        raise ValueError('The block spec values {} do not match with '
                         'the schema {}'.format(block_specs[0], block_spec_schema))

    decoded_specs = []

    for spec in block_specs:
        spec_dict = dict(zip(block_spec_schema, spec))
        decoded_specs.append(spec_dict)

    block_id = 1
    first_layer = False
    for ds in decoded_specs:
        ds['filters'] = utils.make_divisible(
            value=ds['filters'] * width_multiplier,
            divisor=divisible_by,
            min_value=8)
        if block_id >= 5 and reduce_last_block_filters:
            ds['filters'] = ds['filters'] // 2
            if first_layer:
                ds['expand_ratio'] = ds['expand_ratio'] // 2
        block_id = block_id + 1 if ds['is_endpoint'] else block_id
        first_layer = ds['is_endpoint']

    return decoded_specs


# pylint: enable=invalid-name


class MobileNet(tf.keras.Model):
    """Creates a MobileNetV3 family model."""

    def __init__(
            self,
            model_id: str = 'MobileNetV3Small',
            width_multiplier: float = 1.0,
            output_stride: Optional[int] = None,
            min_width: int = 8,
            divisible_by: int = 8,
            regularize_depthwise: bool = False,
            bn_layer: Callable[..., Any] = tf.keras.layers.BatchNormalization,
            conv_kernel_weight_decay: float = 0.0,
            reduce_last_block_filters: bool = False,
            name: str = 'MobilenNetV3'):
        """Initializes a MobileNet V3 model.

        Args:
          model_id: A `str` of MobileNet version. The supported values are
            `MobileNetV3Large`, `MobileNetV3Small`.
          width_multiplier: A `float` of multiplier for the filters (number of
            channels) for all convolution ops. The value must be greater than zero.
            Typical usage will be to set this value in (0, 1) to reduce the number
            of parameters or computation cost of the model.
          output_stride: An `int` that specifies the requested ratio of input to
            output spatial resolution. If not None, then we invoke atrous
            convolution if necessary to prevent the network from reducing the
            spatial resolution of activation maps. The output_stride should be
            divisible by 4.
          min_width: An `int` of minimum width (number of channels) for all
            convolution ops. Enforced when width_multiplier < 1, and not an active
            constraint when width_multiplier >= 1.
          divisible_by: An `int` that ensures all intermediate feature dimensions
            are divisible by this number.
          regularize_depthwise: If True, apply regularization on depthwise conv.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.
          reduce_last_block_filters: A bool indicates whether to reduce the final
            block's filters by a factor of 2.
          name: Model name.

        Raises:
          ValueError: The MobileNet version is not supported.
          ValueError: width_multiplier is not greater than zero.
          ValueError: Output stride must be None or a multiple of 4.
          ValueError: Unknown block type i for layer j.
        """
        if model_id not in SUPPORTED_SPECS_MAP:
            raise ValueError('The MobileNet version {} '
                             'is not supported'.format(model_id))

        if width_multiplier <= 0:
            raise ValueError('width_multiplier is not greater than zero.')

        if (output_stride is not None and
                (output_stride <= 1 or (output_stride > 1 and output_stride % 4))):
            raise ValueError('Output stride must be None or a multiple of 4.')

        super().__init__(name=name)

        self._model_id = model_id
        self._width_multiplier = width_multiplier
        self._min_width = min_width
        self._output_stride = output_stride
        self._divisible_by = divisible_by
        self._regularize_depthwise = regularize_depthwise
        self._bn_layer = bn_layer
        self._conv_kernel_weight_decay = conv_kernel_weight_decay
        self._reduce_last_block_filters = reduce_last_block_filters
        self._blocks = []
        self._endpoint_names = []

        block_specs = SUPPORTED_SPECS_MAP.get(model_id)
        self._decoded_specs = _block_spec_decoder(
            specs=block_specs,
            width_multiplier=self._width_multiplier,
            divisible_by=self._divisible_by,
            reduce_last_block_filters=self._reduce_last_block_filters)

        self._mobilenet_base()

    def _mobilenet_base(self):
        """Builds the base MobileNet architecture."""

        # The current_stride variable keeps track of the output stride of the
        # activations, i.e., the running product of convolution strides up to the
        # current network layer. This allows us to invoke atrous convolution
        # whenever applying the next convolution would result in the activations
        # having output stride larger than the target output_stride.
        current_stride = 1

        # The atrous convolution rate parameter.
        rate = 1

        endpoint_level = 1
        in_filters = _INPUT_CHANNELS
        for i, block_def in enumerate(self._decoded_specs):
            # We only need to build up to 'res5' endpoint for segmentation task.
            if endpoint_level > 5:
                break

            block_name = '{}_{}'.format(block_def['block_fn'], i + 1)

            if (self._output_stride is not None and
                    current_stride == self._output_stride):
                # If we have reached the target output_stride, then we need to employ
                # atrous convolution with stride=1 and multiply the atrous rate by the
                # current unit's stride for use in subsequent layers.
                layer_stride = 1
                layer_rate = rate
                rate = (
                    rate * block_def['strides']
                    if block_def['strides'] is not None else rate)
            else:
                layer_stride = block_def['strides']
                layer_rate = 1
                current_stride = (
                    current_stride * block_def['strides']
                    if block_def['strides'] is not None else current_stride)

            if block_def['block_fn'] == 'conv_bn':

                self._blocks.append(
                    convolutions.Conv2DSame(
                        output_channels=block_def['filters'],
                        kernel_size=block_def['kernel_size'],
                        strides=layer_stride,
                        atrous_rate=layer_rate,
                        activation=block_def['activation'],
                        use_bias=False,
                        bn_layer=self._bn_layer,
                        use_bn=True,
                        conv_kernel_weight_decay=self._conv_kernel_weight_decay,
                        name=block_name,
                    ))

            elif block_def['block_fn'] == 'inverted_bottleneck':
                atrous_rate = 1
                # There is no need to apply atrous convolution to any 1x1 convolution.
                if layer_rate > 1 and block_def['kernel_size'] != 1:
                    atrous_rate = layer_rate
                self._blocks.append(
                    blocks.InvertedBottleneckBlock(
                        in_filters=in_filters,
                        out_filters=block_def['filters'],
                        expand_ratio=block_def['expand_ratio'],
                        strides=layer_stride,
                        kernel_size=block_def['kernel_size'],
                        se_ratio=block_def['se_ratio'],
                        activation=block_def['activation'],
                        expand_se_in_filters=True,
                        depthwise_activation=None,
                        atrous_rate=atrous_rate,
                        divisible_by=self._divisible_by,
                        regularize_depthwise=self._regularize_depthwise,
                        use_depthwise=True,
                        # Note that whether the residual connection would be used is
                        # also conditional on the in_filters and out_filters size, even
                        # if use_residual=True,e.g. when input_filters != out_filters,
                        # no residual connection will be created.
                        use_residual=(block_def['strides'] == 1),
                        bn_layer=self._bn_layer,
                        conv_kernel_weight_decay=self._conv_kernel_weight_decay,
                        name=block_name,
                    ))

            else:
                raise ValueError('Unknown block type {} for layer {}'.format(
                    block_def['block_fn'], i))

            # Register input_filters for the next level
            in_filters = block_def['filters']

            if block_def['is_endpoint']:
                # Name the endpoint to be 'res{1...5}' to align with ResNet. This
                # simplifies segmentation head implementation.
                self._endpoint_names.append('res' + str(endpoint_level))
                endpoint_level += 1
            else:
                self._endpoint_names.append(None)

    def call(self, input_tensor: tf.Tensor, training: bool = False):
        """Performs a forward pass through MobileNet."""
        net = input_tensor
        endpoints = {}
        for block, endpoint_name in zip(self._blocks, self._endpoint_names):
            if isinstance(block, blocks.InvertedBottleneckBlock):
                net, depthwise_output = block(net, training=training)
                if endpoint_name is not None:
                    # Use the corresponding layer's 'depthwise_output' as the endpoint for
                    # segmentation task if possible.
                    endpoints[endpoint_name] = depthwise_output
            else:
                net = block(net, training=training)
                if endpoint_name is not None:
                    endpoints[endpoint_name] = net
        return endpoints


def MobileNetV3Small(
        width_multiplier: float = 1.0,
        output_stride: int = 32,
        bn_layer: Callable[..., Any] = tf.keras.layers.BatchNormalization,
        conv_kernel_weight_decay: float = 0.0,
        reduce_last_block_filters: bool = False,
        name: str = 'MobileNetV3Small') -> tf.keras.Model:
    """Creates a MobileNetV3Small model.

    Args:
      width_multiplier: A float, depth_multiplier for the whole model.
      output_stride: An optional integer specifying the output stride of the
        network.
      bn_layer: An optional tf.keras.layers.Layer that computes the
          normalization (default: tf.keras.layers.BatchNormalization).
      conv_kernel_weight_decay: A float, the weight decay for convolution kernels.
      reduce_last_block_filters: A bool indicates whether to reduce the final
        block's filters by a factor of 2.
      name: Model name.

    Returns:
      The MobileNetV3Small model as an instance of tf.keras.Model.
    """
    model = MobileNet(model_id='MobileNetV3Small',
                      width_multiplier=width_multiplier,
                      output_stride=output_stride,
                      bn_layer=bn_layer,
                      conv_kernel_weight_decay=conv_kernel_weight_decay,
                      reduce_last_block_filters=reduce_last_block_filters,
                      name=name)
    return model


def MobileNetV3Large(
        width_multiplier: float = 1.0,
        output_stride: int = 32,
        bn_layer: Callable[..., Any] = tf.keras.layers.BatchNormalization,
        conv_kernel_weight_decay: float = 0.0,
        reduce_last_block_filters: bool = False,
        name: str = 'MobileNetV3Large') -> tf.keras.Model:
    """Creates a MobileNetV3Large model.

    Args:
      width_multiplier: A float, depth_multiplier for the STEM.
      output_stride: An optional integer specifying the output stride of the
        network.
      bn_layer: An optional tf.keras.layers.Layer that computes the
          normalization (default: tf.keras.layers.BatchNormalization).
      conv_kernel_weight_decay: A float, the weight decay for convolution kernels.
      reduce_last_block_filters: A bool indicates whether to reduce the final
        block's filters by a factor of 2.
      name: Model name.

    Returns:
      The MobileNetV3Large model as an instance of tf.keras.Model.
    """
    model = MobileNet(model_id='MobileNetV3Large',
                      width_multiplier=width_multiplier,
                      output_stride=output_stride,
                      bn_layer=bn_layer,
                      conv_kernel_weight_decay=conv_kernel_weight_decay,
                      reduce_last_block_filters=reduce_last_block_filters,
                      name=name)
    return model



#!/usr/bin/env python3
r"""

SeeAlso:
    ~/code/watch/geowatch/cli/queue_cli/prepare_time_combined_dataset.py


CommandLine:
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)

    python -m geowatch.cli.coco_time_combine \
        --input_kwcoco_fpath="$DVC_DATA_DPATH/Drop6/imgonly-KR_R002.kwcoco.zip" \
        --output_kwcoco_fpath="$DVC_DATA_DPATH/Drop6_MeanYear/imgonly-KR_R002.kwcoco.zip" \
        --channels="red|green|blue|nir|swir16|swir22" \
        --resolution=10GSD \
        --time_window=1year \
        --merge_method=mean \
        --workers=4

    python -m geowatch reproject_annotations \
        --src $DVC_DATA_DPATH/Drop6_MeanYear/imgonly-KR_R002.kwcoco.zip \
        --dst $DVC_DATA_DPATH/Drop6_MeanYear/imganns-KR_R002.kwcoco.zip \
        --site_models="$DVC_DATA_DPATH/annotations/drop6/site_models/*.geojson"


Ignore:
    # Debugging

    python -m geowatch.cli.coco_time_combine \
        --kwcoco_fpath="$HOME/data/dvc-repos/smart_data_dvc/Aligned-Drop7/VN_C002/imgonly-VN_C002-rawbands.kwcoco.zip" \
        --output_kwcoco_fpath="$HOME/data/dvc-repos/smart_data_dvc/Drop7-MedianNoWinter10GSD-V2/VN_C002/_unfielded_imgonly-VN_C002-rawbands.kwcoco.zip" \
        --channels="red|green|blue|nir|swir16|swir22|pan|coastal|cirrus|B05|B06|B07|B8A|B09" \
        --resolution="10GSD" \
        --time_window=1y \
        --remove_seasons=winter \
        --merge_method=median \
        --spatial_tile_size=1024 \
        --mask_low_quality=True \
        --start_time=2010-03-01 \
        --assets_dname="raw_bands" \
        --workers=0


Example:
    >>> # Toydata example for CI
    >>> import geowatch
    >>> from geowatch.cli import coco_time_combine
    >>> import ubelt as ub
    >>> dpath = ub.Path.appdir('geowatch/tests/cli/time_combine/t0')
    >>> dset = geowatch.coerce_kwcoco(
    >>>     'geowatch-msi', geodata=True,
    >>>     dates={'start_time': '2020-01-01', 'end_time': '2020-06-01'},
    >>>     image_size=(32, 32)
    >>> )
    >>> dpath.delete().ensuredir()
    >>> output_fpath = dpath / 'time_combined/data.kwcoco.json'
    >>> gsd = dset.videos().objs[0]['target_gsd']
    >>> kwargs = coco_time_combine.TimeCombineConfig(
    >>>     input_kwcoco_fpath=dset.fpath,
    >>>     output_kwcoco_fpath=output_fpath,
    >>>     time_window='2month',
    >>>     merge_method='mean',
    >>>     resolution=f'{gsd}GSD',
    >>>     start_time='2019-06-01',
    >>> )
    >>> cmdline = 0
    >>> coco_time_combine.main(cmdline=cmdline, **kwargs)
    >>> import kwcoco
    >>> out_dset = kwcoco.CocoDataset(output_fpath)
    >>> assert len(out_dset.videos()) == len(dset.videos())
    >>> assert out_dset.n_images < dset.n_images

    from geowatch.cli import coco_visualize_videos
    coco_visualize_videos.main(cmdline=0, src=output_fpath, stack='only', workers='avail')

"""
import os
import ubelt as ub
import scriptconfig as scfg

try:
    from line_profiler import profile
except Exception:
    profile = ub.identity


class TimeCombineConfig(scfg.DataConfig):
    """
    Averages kwcoco images over a sliding temporal window in a video.
    """
    __command__ = 'time_combine'

    input_kwcoco_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        The path to the kwcoco file containing the image data to be
        combined.
        '''), alias=['kwcoco_fpath'], position=1)

    output_kwcoco_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        The path where the combined image data will be saved to in a
        kwcoco file.
        '''), position=2)

    channels = scfg.Value('*', help=ub.paragraph(
        '''
        The channels to get and combine the spatial data from. E.g.
        "red|green|blue". Note: Separate channels with "|". A ``*``
        means combine everything.
        '''))

    time_window = scfg.Value('1month', help=ub.paragraph(
        '''
        The temporal window that will group images. E.g. 1y (or 365d) will
        group all images in every year. Buckets are non-overlapping.  The first
        image defines where the buckets start.
        '''), alias=['temporal_window_duration'])

    start_time = scfg.Value(None, help=ub.paragraph(
        '''
        The datetime to start time window partitioning. If unspecified the
        first date in the dataset is used.
        '''))

    merge_method = scfg.Value('mean', help=ub.paragraph(
        '''
        How to combine multiple observations over each time_window.
        '''), choices=['mean', 'median', 'max'])

    resolution = scfg.Value('10GSD', help=ub.paragraph(
        '''
        The resolution the imagery will be loaded during the
        combination operation and saved to the output kwcoco file.
            '''))

    mask_low_quality = scfg.Value(True, isflag=True, help=ub.paragraph(
        '''
        If True, use the quality masks to prevent low-quality pixels from
        being used in the average.
        '''), alias=['filter_with_cloudmasks'])

    sensor_weights = scfg.Value(0, help=ub.paragraph(
        '''
        Weight the contribution of each sensor in the average operation
        by a scalar value.
        '''))

    separate_sensors = scfg.Value(True, isflag=True, help=ub.paragraph(
        '''
        Combine images by sensor separately, otherwise bands from multiple
        sensors are averaged together.
        '''))

    workers = scfg.Value(0, help=ub.paragraph(
            '''
            The number of CPU cores to compute the combination operation with.
            '''))

    include_sensors = scfg.Value(None, help=ub.paragraph(
            '''
            A list of sensors to include in the combination operation.
            '''))

    assets_dname = scfg.Value('_assets', help=ub.paragraph(
        '''
        The name of the top-level directory to write new assets.
        '''))

    exclude_sensors = scfg.Value(None, help=ub.paragraph(
        '''
        A list of sensors to exclude from the combination operation.
        '''))

    select_images = scfg.Value(None, help='see kwcoco_extensions.filter_image_ids docs')

    select_videos = scfg.Value(None, help='see kwcoco_extensions.filter_image_ids docs')

    spatial_tile_size = scfg.Value(None, help=ub.paragraph(
        '''
        The size of the tiling over space to use when computing the average.
        If None, the average is computed on the entire stack of images at
        once.
        '''))

    filter_season = scfg.Value([], nargs='+', help=ub.paragraph(
        '''
        Images within this season(s) will be excluded from the average operation.
        Season options: ['spring', 'summer', 'fall', 'winter']
        '''), alias=['remove_seasons'])

    max_images_per_group = scfg.Value(None, help=ub.paragraph(
        '''
        If specified, the averaging operation (e.g. mean/median) will only
        consider a subset of the images within each temporal window.  This can
        greatly reduce the resources required to run this script at the cost of
        quality. Currently a heuristic is used to select the "highest quality"
        subset of images.
        '''))

    avoid_quality_values = scfg.Value(['cloud'], nargs='+', help=ub.paragraph(
        '''
        The types of QA categories we will avoid.
        '''))
    # avoid_quality_values = ['cloud', 'cloud_shadow', 'cloud_adjacent']
    # avoid_quality_values += ['ice']


def main(cmdline=1, **kwargs):
    """
    CommandLine:
        DEVEL_TEST=1 xdoctest -m geowatch.cli.coco_time_combine main

        from geowatch.cli.coco_time_combine import *  # NOQA
        import geowatch
        data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        cmdline = 0
        channels='red|green|blue'
        kwargs = dict(
            input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imgonly-KR_R001.kwcoco.zip',
            output_kwcoco_fpath=data_dvc_dpath / 'TestAveDrop6/test-timeave-valid_split1_1yr_mean_test.kwcoco.zip',
            workers=4,
            mask_low_quality=True,
            time_window='1 year',
            channels=channels,
        )
        output_coco_dset = main(cmdline=cmdline, **kwargs)
        coco_visualize_videos.main(cmdline=cmdline, src=kwargs['output_kwcoco_fpath'], smart=True)

    Example:
        >>> # 0: Baseline run.
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imgonly-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-imgonly-KR_R001.kwcoco.zip',
        >>>     workers=11,
        >>>     mask_low_quality=False,
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-imgonly-KR_R001.kwcoco.zip

    Example:
        >>> # 1: Check cloudmasking.
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imgonly-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-KR_R001-cloudmask.kwcoco.zip',
        >>>     workers=11,
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=True,
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001.kwcoco.zip

    Example:
        >>> # 2: Check that resolution can be updated.
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imgonly-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-KR_R001-cloudmask-5GSD.kwcoco.zip',
        >>>     workers=11,
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=True,
        >>>     resolution='5GSD',
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001-cloudmask.kwcoco.zip

    Example:
        >>> # 3: Median combining.
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imgonly-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-KR_R001-median.kwcoco.zip',
        >>>     workers=11,
        >>>     merge_method='median',
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=False,
        >>>     resolution='10GSD',
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001-median.kwcoco.zip

    Example:
        >>> # 4: Median combining with cloudmask.
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imganns-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-KR_R001-cloudmask-median.kwcoco.zip',
        >>>     workers=11,
        >>>     merge_method='median',
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=True,
        >>>     resolution='10GSD',
        >>>     separate_sensors=True,
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001-cloudmask-median.kwcoco.zip

    Example:
        >>> # 5: Dont separate sensors.
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imgonly-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-KR_R001-cloudmask-no_sensor_separate.kwcoco.zip',
        >>>     workers=11,
        >>>     merge_method='mean',
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=True,
        >>>     resolution='10GSD',
        >>>     separate_sensors=False,
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001-cloudmask-no_sensor_separate.kwcoco.zip

    Example:
        >>> # 6: Adjust the effect of S2 imagery.
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imganns-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-test_6-KR_R001.kwcoco.zip',
        >>>     workers=11,
        >>>     merge_method='mean',
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=True,
        >>>     resolution='10GSD',
        >>>     separate_sensors=False,
        >>>     sensor_weights=dict(S2=2.5, L8=1.0, WV=5.0)
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001-cloudmask-s2w_10.kwcoco.zip

    Example:
        >>> # 7: Tile images instead of computing average all at once.
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imganns-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-test_7-KR_R001.kwcoco.zip',
        >>>     workers=11,
        >>>     merge_method='median',
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=True,
        >>>     resolution='10GSD',
        >>>     separate_sensors=False,
        >>>     spatial_tile_size=200,
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001-cloudmask-s2w_10.kwcoco.zip

    Example:
        >>> # 8: Tile images on large kwcoco file.
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/data_train_split1.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-test_8-train_split1.kwcoco.zip',
        >>>     workers=4,
        >>>     merge_method='median',
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=True,
        >>>     resolution='10GSD',
        >>>     separate_sensors=False,
        >>>     spatial_tile_size=400,
        >>>     include_sensors=['S2', 'L8'],
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001-cloudmask-s2w_10.kwcoco.zip

    Example:
        >>> # 9: Exclude winter seasons for time average.
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> cmdline = 0
        >>> kwargs = dict(
        >>>     input_kwcoco_fpath=data_dvc_dpath / 'Drop6/imganns-KR_R001.kwcoco.zip',
        >>>     output_kwcoco_fpath=data_dvc_dpath / 'Drop6/test-timeave-test_9-KR_R001.kwcoco.zip',
        >>>     workers=11,
        >>>     merge_method='mean',
        >>>     time_window='1 year',
        >>>     channels='red|green|blue',
        >>>     mask_low_quality=True,
        >>>     resolution='10GSD',
        >>>     separate_sensors=False,
        >>>     filter_season='winter',
        >>> )
        >>> output_coco_dset = main(cmdline=cmdline, **kwargs)
        >>> from geowatch.cli import coco_visualize_videos
        >>> coco_visualize_videos.main(cmdline=0, src=kwargs['output_kwcoco_fpath'], smart=True)

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        geowatch stats $DVC_DATA_DPATH/Drop6/test-timeave-KR_R001-cloudmask-s2w_10.kwcoco.zip

    """
    config = TimeCombineConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    rich.print('config = ' + ub.urepr(config, nl=1))

    output_coco_dset = combine_kwcoco_channels_temporally(config)
    return output_coco_dset


@profile
def combine_kwcoco_channels_temporally(config):
    """Combine spatial data within a temporal window from a kwcoco dataset and save the result to a new kwcoco dataset.

    High level steps:
    1. Load kwcoco dataset.
    2. Divide the dataset into temporal windows.
    3. For each temporal window, combine the spatial data from each channel.
    4. Save the combined image result to a new kwcoco dataset.
    """
    import kwarray
    import kwcoco
    import numpy as np
    from kwutil import util_progress
    from kwutil import util_time
    from geowatch.utils import kwcoco_extensions
    from kwutil import util_parallel
    from kwutil.util_yaml import Yaml
    from geowatch.utils import util_hardware

    # combining over time seems to only make sense in a video context
    space = 'video'
    workers = util_parallel.coerce_num_workers(config.workers)

    ## Check input kwcoco file path exists.
    if os.path.exists(config.input_kwcoco_fpath) is False:
        raise FileNotFoundError(f'Input kwcoco file path does not exist: {config.input_kwcoco_fpath}')

    sensor_weights = Yaml.coerce(config.sensor_weights)

    ## Check the S2 weight factor and merge method combination.
    if isinstance(sensor_weights, dict) is False:
        sensor_weights = {}

    for sensor_code, sensor_weight in sensor_weights.items():
        if sensor_weight != 1.0 and config.merge_method != 'mean':
            print(f'WARNING: {sensor_code} weight factor only effects the merge method "mean".')

    # 1. Load kwcoco dataset.
    coco_dset = kwcoco.CocoDataset.coerce(config.input_kwcoco_fpath)
    print(f'coco_dset = {ub.urepr(coco_dset, nl=1)}')

    if coco_dset.n_images == 0:
        raise ValueError('Input kwcoco dataset contains no images')

    selected_gids = kwcoco_extensions.filter_image_ids(
        coco_dset,
        include_sensors=config['include_sensors'],
        exclude_sensors=config['exclude_sensors'],
        select_images=config['select_images'],
        select_videos=config['select_videos'],
    )

    # Optional: Filter image IDs by season.
    if config.filter_season is not None:
        selected_gids = filter_image_ids_by_season(coco_dset, selected_gids, config.filter_season)

    if len(selected_gids) == 0:
        raise ValueError('Input kwcoco dataset contained images, but no images were selected')

    if selected_gids is not None:
        coco_dset = coco_dset.subset(selected_gids)

    output_coco_dset = coco_dset.copy()
    output_coco_dset.clear_images()  # we will write all new images

    requested_sensorchan = kwcoco.SensorChanSpec.coerce(config.channels)
    requested_chans = requested_sensorchan.chans

    ## Get saved asset directory.
    output_kwcoco_fpath = ub.Path(config.output_kwcoco_fpath)
    new_bundle_dpath = output_kwcoco_fpath.parent
    new_bundle_dpath.ensuredir()

    output_coco_dset.fpath = output_kwcoco_fpath

    # 2. Divide the dataset into temporal windows (per video).

    ## Convert time_window from days to seconds.
    time_delta = util_time.coerce_timedelta(config.time_window)
    start_time = util_time.coerce_datetime(config.start_time)
    time_delta_seconds = time_delta.total_seconds()

    video_ids = list(coco_dset.videos())

    pman = util_progress.ProgressManager()
    # pman = util_progress.ProgressManager(backend='progiter')  # uncomment if you need to use breakpoints

    resolution = config.resolution
    merge_method = config.merge_method
    mask_low_quality = config.mask_low_quality
    og_kwcoco_fpath = config.input_kwcoco_fpath
    if isinstance(config.spatial_tile_size, tuple) is False:
        spatial_tile_size = (config.spatial_tile_size, config.spatial_tile_size)
    else:
        spatial_tile_size = config.spatial_tile_size

    n_combined_images = 0
    n_failed_merges = 0
    max_percent = 0

    with pman:

        for video_id in pman.progiter(video_ids, desc='Average across time for each video'):
            # Get all image ids for the video.
            images = coco_dset.images(video_id=video_id)

            # Determine what channels in each image we want to merge
            image_merge_channels = []
            for coco_img in images.coco_images:
                if requested_chans.spec == '*':
                    # Merge everything
                    mergable = coco_img.channels.fuse().to_set()
                else:
                    mergable = requested_chans.to_set() & coco_img.channels.fuse().to_set()
                image_merge_channels.append(frozenset(mergable))

            video = coco_dset.index.videos[video_id]
            video_name = video['name']
            video_dsize = (video['width'], video['height'])
            print(f'video_dsize={video_dsize}')

            # Filter to only images that have a channel to merge
            flags = [len(c) > 0 for c in image_merge_channels]
            image_merge_channels = list(ub.compress(image_merge_channels, flags))
            images = images.compress(flags)

            image_datetimes = list(map(util_time.coerce_datetime, images.lookup('timestamp')))

            image_unixtimes = np.array([t.timestamp() for t in image_datetimes])

            # Get number of seconds after the start time (or first image)
            if start_time is not None:
                image_rel_unixtimes = image_unixtimes - start_time.timestamp()
            else:
                image_rel_unixtimes = image_unixtimes - image_unixtimes[0]

            # We might group via multiple criteria
            groupers = {}

            # Assign each image to a temporal bucket/window
            image_buckets = np.floor(image_rel_unixtimes / time_delta_seconds).astype(int)
            groupers['bucket'] = image_buckets

            # groupers['channels'] = image_merge_channels  # not sure if we do this or not

            if config.separate_sensors:
                image_sensors = images.lookup('sensor_coarse', None)
                groupers['sensor'] = image_sensors

            # Construct a group-id for each image
            image_groupids = list(zip(*groupers.values()))

            # Group all image-indexes within the same group-id together.
            # unique_groupids, grouped_idxs = kwarray.group_indices(image_groupids)
            # groupid_to_idxs = dict(zip(unique_groupids, grouped_idxs))
            groupid_to_idxs = ub.group_items(range(len(image_groupids)), image_groupids)

            print(f'{len(images)} / {len(flags)} images in {video_name} have {len(groupid_to_idxs)} mergable groups')

            SHOW_INFO = 1
            # DEBUG: Print the distribution of images per window.
            if SHOW_INFO:
                # Get the histogram for the number of images per window.
                bucket_to_num_images = ub.udict(groupid_to_idxs).map_values(len)
                bucket_stats = kwarray.stats_dict(list(bucket_to_num_images.values()))
                print(f'[{video_name}] Distribution of images per window:')
                print('Histogram: = {}'.format(ub.urepr(bucket_to_num_images, nl=1)))
                print('N images per window: = {}'.format(ub.urepr(bucket_stats, nl=1)))

            jobs = ub.JobPool(mode='process', max_workers=workers)

            with jobs:
                # 3. For each temporal window, combine the spatial data from each channel.
                chunk_image_idxs = list(groupid_to_idxs.values())
                prog = pman.progiter(chunk_image_idxs, transient=True, desc='Submit combine within temporal windows jobs')
                for chunk_image_idxs in prog:
                    # if len(chunk_image_idxs) == 1:
                    #     import xdev
                    #     xdev.embed()
                    # else:
                    #     continue
                    window_images = images.take(chunk_image_idxs)
                    window_coco_images = window_images.coco_images
                    # Detach for process parallelization
                    window_coco_images = [g.detach() for g in window_coco_images]

                    job = jobs.submit(merge_images, window_coco_images,
                                      merge_method, requested_chans, space,
                                      resolution, new_bundle_dpath,
                                      mask_low_quality, sensor_weights,
                                      og_kwcoco_fpath, spatial_tile_size,
                                      config=config)
                    job.merge_images = merge_images

                for job in pman.progiter(jobs.as_completed(),
                                         total=len(jobs),
                                         desc='Collect combine within temporal windows jobs'):
                    new_img = job.result()
                    if new_img is None:
                        n_failed_merges += 1
                        continue
                    output_coco_dset.add_image(**new_img)
                    n_combined_images += 1

                    mem_info = util_hardware.get_mem_info()
                    curr_percent = mem_info['percent']
                    max_percent = max(max_percent, curr_percent)
                    pman.update_info(ub.codeblock(
                        f'''
                        {n_combined_images=}
                        {n_failed_merges=}
                        Memory Percent: {curr_percent}%
                        Max Percent: {max_percent}%
                        '''))

    if n_combined_images == 0:
        raise ValueError('No images were combined with non-NaN values')

    # from geowatch.utils import util_resolution
    from kwcoco.coco_image import coerce_resolution
    target_gsd = float(np.mean(coerce_resolution(config.resolution)['mag']))
    print(f'Reset geowatch feilds target_gsd={target_gsd}')
    kwcoco_extensions.populate_watch_fields(
        output_coco_dset, target_gsd=target_gsd, overwrite=True)

    # video = output_coco_dset.index.videos[video_id]
    # after_video_dsize = (video['width'], video['height'])
    # print(f'after_video_dsize={after_video_dsize}')

    # for video_id in ub.ProgIter(vidids, total=len(vidids), desc='populate videos'):
    #     coco_populate_geo_video_stats(coco_dset, video_id, target_gsd=target_gsd)

    # Debugging:
    if 0:
        kwcoco_extensions.check_kwcoco_spatial_transforms(output_coco_dset)

    # Save kwcoco file.
    print(f"Saving ouput kwcoco file to: {output_kwcoco_fpath}")
    if 0:
        output_coco_dset.validate()
    output_coco_dset.dump()
    print(f"Saved ouput kwcoco file to: {output_kwcoco_fpath}")
    return output_coco_dset


@profile
def get_quality_mask(coco_image, space, resolution, avoid_quality_values=None, crop_slice=None, qa_scale=None):
    """
    Get a binary mask of the quality data.

    Args:
        coco_image (kwcoco.coco_image.CocoImage):
            Object that contains references to the image and assets including the quality mask.

        space (str):
            The space that the quality mask will be loaded in. Choices: 'image', 'video', 'asset'

        resolution (str | int):
            The resolution that the quality mask will be loaded in. E.g. '10GSD'.

        avoid_quality_values (list | None):
            The values to include as bad quality according to the bitmask. Defaults to ['cloud', 'cloud_shadow', 'cloud_adjacent'].

        crop_slice (Tuple[slice, slice] | None):
            The height and width crop slices to load from quality mask. Defaults to None, which loads full quality mask.

    Returns:
        np.ndarray: A binary numpy array of shape [H, W, 1] where the 1 values corresponds to a quality pixel vice versa for 0 values.
    """
    from geowatch.tasks.fusion.datamodules.qa_bands import QA_SPECS
    from geowatch import heuristics
    import numpy as np
    # spec_name = coco_image.img.get('qa_encoding', heuristics.DEFAULT_QA_ENCODING)
    # sensor = coco_image.img.get('sensor_coarse', '*')

    delay = coco_image.imdelay('quality',
                               space=space,
                               interpolation='nearest',
                               antialias=False,
                               resolution=resolution)
    if crop_slice:
        delay = delay.crop(crop_slice)
    if qa_scale:
        delay = delay.scale(qa_scale)

    qa_asset = coco_image.find_asset('quality')
    if qa_asset is None:
        # No quality mask is available, don't finalize, return 1s.
        quality_mask = np.ones(delay.shape, dtype=np.uint8)
        return quality_mask

    qa_data = delay.finalize(antialias=False, interpolation='nearest',
                             nodata_method='ma')
    if hasattr(qa_data, 'mask'):
        is_nodata = qa_data.mask
    else:
        is_nodata = None
        import warnings
        warnings.warn(
            'Warning: nodata_method was ma, but we didnt get a mask back. '
            'This is likely a delayed-image issue')

    spec_name = qa_asset.get('qa_encoding', heuristics.DEFAULT_QA_ENCODING)
    sensor = qa_asset.get('sensor_coarse', '*')
    try:
        table = QA_SPECS.find_table(spec_name, sensor)
    except AssertionError as ex:
        print(f'warning ex={ex}')
        is_iffy = None
    else:
        is_iffy = table.mask_any(qa_data.data, avoid_quality_values)
    if is_nodata is not None:
        is_iffy |= is_nodata

    if is_iffy is None:
        quality_mask = np.ones(delay.shape, dtype=np.uint8)
    else:
        quality_mask = (1 - is_iffy)
    return quality_mask


@profile
def estimate_top_quality_images(window_coco_images, max_images_per_group, avoid_quality_values):
    """
    Heuristic to subselect top images.
    """
    import numpy as np

    assert len(window_coco_images) > max_images_per_group

    qa_assets = [coco_image.find_asset('quality')
                 for coco_image in window_coco_images]
    has_known_encoding = [
        qa_asset is not None and 'qa_encoding' in qa_asset
        for qa_asset in qa_assets
    ]
    # We can try to use QA bands if we have those
    use_qa_bands = all(has_known_encoding)
    estimated_quality = []

    # Use an overview to estimate quality
    qa_scale = 0.25

    if use_qa_bands:
        for coco_image in window_coco_images:
            # Note: might want to configure the resolution we are loading at
            quality_mask = get_quality_mask(
                coco_image, space='video', resolution=None,
                avoid_quality_values=avoid_quality_values, qa_scale=qa_scale,
                crop_slice=None)

            quality = quality_mask.sum() / quality_mask.size
            estimated_quality.append(quality)
    else:
        # If we have no information on QA bands, we can use the coarse stac
        # properties of the larger images
        inf = float('inf')
        for coco_image in window_coco_images:
            # FIXME, non general hard-coded properties used here
            ave_cloudcover = np.nanmean([
                prop.get('eo:cloud_cover', inf) / 100
                for prop in coco_image.img.get('parent_stac_properties', [])
            ])
            estimated_quality.append(1 - ave_cloudcover)

    ranked_idxs = np.argsort(estimated_quality)[::-1]
    top_idxs = ranked_idxs[:max_images_per_group]
    chosen = list(ub.take(window_coco_images, top_idxs))
    chosen_window_coco_images = sorted(chosen, key=lambda x: x.datetime)
    return chosen_window_coco_images


@profile
def merge_images(window_coco_images, merge_method, requested_chans, space,
                 resolution, new_bundle_dpath, mask_low_quality,
                 sensor_weights, og_kwcoco_fpath, spatial_tile_size, config):
    """
    Args:
        window_coco_images (List[kwcoco.CocoImage]): images with channels to merge

    Returns:
        Dict: a new coco image that points at the merged image on disk.
    """
    import kwimage
    import kwarray
    import kwcoco
    import numpy as np
    from geowatch.tasks.fusion.coco_stitcher import CocoStitchingManager
    from kwutil import util_time

    # Determine what channels are available in this image set
    available_chans_set = set(ub.flatten([g.channels.fuse().to_set() for g in window_coco_images]))

    if requested_chans.spec == '*':
        requested_chans_set = available_chans_set
    else:
        requested_chans_set = requested_chans

    merge_chans_set = requested_chans_set & available_chans_set
    merge_chans = kwcoco.FusedChannelSpec.coerce(sorted(merge_chans_set))

    # Take only the N "best" images per group to combine over.
    # By default we use a hard coded property to identify cloud cover
    # TODO: make this more general with a standardized kwcoco coarse
    # quality property
    max_images_per_group = config.max_images_per_group
    if max_images_per_group is not None and len(window_coco_images) > max_images_per_group:
        window_coco_images = estimate_top_quality_images(
            window_coco_images, max_images_per_group,
            config.avoid_quality_values)

    # TODO: we should merge each asset at the highest resolution for that
    # asset, but no more.  For instance, when given red|green|blue|swir16 we
    # should handle rgb separately from swir16. This will involve grouping by
    # the asset resolution of each image. We need to maintain video space
    # alignment, but we can adjust by scalefactors. For now we are ignoring all
    # of this.

    # For now we just choose video space
    assert space == 'video'

    first_coco_img = window_coco_images[0]

    video = first_coco_img.video
    video_name = video['name']

    # Scales to the resolution from the requested (i.e. video space)
    scale_asset_from_vid = first_coco_img._scalefactor_for_resolution(resolution=resolution, space='video')
    # warp_asset_from_vid = kwimage.Affine.scale(scale_asset_from_vid)

    video_dsize = kwimage.Box.from_dsize((video['width'], video['height']))

    canvas_dsize = video_dsize.scale(scale_asset_from_vid).quantize().dsize
    canvas_dims = canvas_dsize[::-1]
    video_height, video_width = canvas_dims
    canvas_shape = canvas_dims + (merge_chans.numel(), )

    if all(spatial_tile_size) is False:
        # Set tiles as the same size as the video.
        slider = [(slice(0, video_height, None), slice(0, video_width, None))]
    else:
        # Create tiling slices for given video shape and tile size.

        ## Check that crop size is not bigger than the video size.
        if video_height < spatial_tile_size[0]:
            spatial_height_size = video_height
        else:
            spatial_height_size = spatial_tile_size[0]

        if video_width < spatial_tile_size[1]:
            spatial_width_size = video_height
        else:
            spatial_width_size = spatial_tile_size[1]

        # step = min(spatial_height_size, spatial_width_size)

        slider = kwarray.SlidingWindow(
            shape=(video_height, video_width),
            window=(spatial_height_size, spatial_width_size),
            overlap=0,
            keepbound=True,
            allow_overshoot=True
        )

    avoid_quality_values = config.avoid_quality_values
    # Create canvas to combine averaged tiles into.

    # todo: memmap this.
    average_canvas = kwarray.Stitcher(canvas_shape)

    import warnings
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'All-NaN')
        warnings.filterwarnings('ignore', 'invalid')

        # Load and combine the images within this range.
        for crop_slice in slider:

            frame_gen = generate_frames(window_coco_images, crop_slice,
                                        resolution, space, merge_chans,
                                        mask_low_quality, avoid_quality_values)

            if merge_method == 'mean':
                accum = kwarray.Stitcher(canvas_shape)

                # TODO: we can also parallelize this in case the window is really big
                for coco_img, image_data in frame_gen:
                    pxl_weight = (~np.isnan(image_data)).astype(np.float32)
                    try:
                        sensor_weight = sensor_weights[coco_img['sensor_coarse']]
                    except KeyError:
                        sensor_weight = 1.0
                    pxl_weight *= sensor_weight
                    image_data = np.nan_to_num(image_data)
                    accum.add((slice(None), slice(None)), image_data, pxl_weight)

                combined_image_data = accum.finalize()

            elif merge_method == 'remedian':
                # from remedian.remedian import Remedian
                from geowatch.utils.remedian import Remedian
                num_frames = len(window_coco_images)
                _, first_frame = next(frame_gen)
                data_shape = first_frame.shape
                approx_median = Remedian(data_shape, n_obs=7, t=num_frames, allow_nan=True)
                approx_median.add_obs(first_frame)
                del first_frame
                for _, image_data in frame_gen:
                    approx_median.add_obs(image_data)
                combined_image_data = approx_median.remedian

            elif merge_method == 'median':
                # TODO: Make this less computationally expensive.
                median_stack = [t[1] for t in frame_gen]
                combined_image_data = np.nanmedian(median_stack, axis=0, overwrite_input=True)
                # TODO: Fix the logic below to match above because it should be faster.
                # matched_quality_mask = np.repeat(quality_mask, repeats=3, axis=2)
                # masked_image_data = np.ma.masked_array(data=image_data2, mask=~matched_quality_mask, fill_value=np.nan)
                # image_data = M.filled(np.nan)
                # masked_image_data = M.filled(np.nan)
                # _median_stack = np.stack(median_stack)
                # combined_image_data = np.nanmedian(_median_stack, axis=0)

            elif merge_method == 'max':
                # TODO: Combine with other methods.
                stack = []
                for coco_img, image_data in frame_gen:
                    stack.append(image_data)
                combined_image_data = np.nanmax(stack, axis=0)
            else:
                raise NotImplementedError

            # Add the combined image data to the average canvas.
            average_canvas.add(crop_slice, combined_image_data)

        # Rename the average canvas to the combined image data.
        combined_image_data = average_canvas.finalize()

    # Check if the image contains data after cloud masking.
    if np.all(np.isnan(combined_image_data)):
        return None

    # 4. Save the combined image result to a new kwcoco dataset.
    ## Use the first image as the standin for the new image.

    # Create a dictionary for the new image
    # TODO: Should likely update properties to indicate what went into this new
    # image.
    new_img = first_coco_img.img.copy()
    new_img.pop('auxiliary', None)
    new_img.pop('assets', None)
    new_img.pop('name', None)
    new_img.pop('align_method', None)
    new_img.pop('valid_region', None)
    new_img.pop('valid_region_utm', None)
    new_img.pop('parent_name', None)
    new_img.pop('parent_canonical_name', None)
    new_coco_img = kwcoco.CocoImage(new_img)

    # We are currently writing all merged assets in video space, so
    # the transform from asset space to image space is the transform
    # from video to image space.
    # warp_vid_from_asset = warp_asset_from_vid.inv()
    # new_warp_img_from_asset = first_coco_img.warp_img_from_vid @ warp_vid_from_asset

    # Create the same image name.
    hash_depends = {
        'names': [g.img['name'] for g in window_coco_images],
        'chans': merge_chans.spec,
        'resolution': resolution,
        'space': space,
        'version': 0,
        'og_coco_path': og_kwcoco_fpath,
        'merge_method': merge_method,
    }
    hashid = ub.hash_data(hash_depends)[0:16]

    timestr = util_time.isoformat(util_time.coerce_datetime(new_coco_img.img['timestamp']))
    chanstr = merge_chans.path_sanitize()
    sensors = '_'.join(sorted(set([coco_img['sensor_coarse'] for coco_img in window_coco_images])))
    new_name = f'ave_{timestr}_{len(window_coco_images):03d}_{chanstr}_{sensors}_{hashid}'
    new_coco_img.img['name'] = new_name

    # TODO: Figure out how to better handle this case.
    # Issue is that the unique sensor combination does not get processed in the predict script.
    new_coco_img.img['sensor_coarse'] = first_coco_img['sensor_coarse']
    new_coco_img.img['_parent_sensors'] = sensors
    new_coco_img.img['_num_parents'] = len(window_coco_images)
    new_coco_img.img['_first_timestamp'] = window_coco_images[0].img['timestamp']
    new_coco_img.img['_last_timestamp'] = window_coco_images[-1].img['timestamp']

    # new_coco_img.img['sensor_coarse'] = '_'.join(
    #     sorted(set([coco_img['sensor_coarse'] for coco_img in window_coco_images])))

    # new_coco_img.bundle_dapth = new_bundle_dpath
    dname = f'{video_name}/ave_{merge_chans.path_sanitize()}'

    # TODO:
    # We could recompute the valid_region and valid_region_utm here
    # as the union of the input items' properties
    new_coco_img.img.pop('valid_region_utm', None)

    # TODO: Figure out how to add geo-metadata to the new image from previous images.
    tmp_dset = kwcoco.CocoDataset()
    tmp_dset.bundle_dpath = new_bundle_dpath
    tmp_dset.add_video(**first_coco_img.video)
    tmp_dset.add_image(**new_coco_img.img)
    stitch_manager = CocoStitchingManager(
        tmp_dset,
        short_code=dname,
        chan_code=merge_chans.spec,
        stiching_space='video',
        write_prediction_attrs=False,
        assets_dname=config.assets_dname,
    )

    gid = new_coco_img.img['id']
    stitch_manager.accumulate_image(gid,
                                    None,
                                    combined_image_data,
                                    asset_dsize=canvas_dsize,
                                    scale_asset_from_stitchspace=scale_asset_from_vid)
    stitch_manager.finalize_image(gid)
    final_img = tmp_dset.imgs[gid]

    return final_img


@profile
def generate_frames(window_coco_images, crop_slice, resolution, space, merge_chans, mask_low_quality, avoid_quality_values):
    import numpy as np
    for coco_img in window_coco_images:
        delayed = coco_img.imdelay(merge_chans, space=space, resolution=resolution)
        delayed = delayed.crop(crop_slice)
        image_data = delayed.finalize(nodata_method='float')

        if mask_low_quality:
            # Load quality mask.
            quality_mask = get_quality_mask(coco_img, space, resolution, avoid_quality_values=avoid_quality_values, crop_slice=crop_slice)

            # Update pixel weights based on quality pixel values.
            x, y = np.where(quality_mask[..., 0] == 0)
            image_data[x, y, :] = np.nan
        yield coco_img, image_data


@profile
def filter_image_ids_by_season(coco_dset, image_ids, filtered_seasons, ignore_winter_torrid_zone=True):
    """Filter a sequence of image ids by season and geolocation.

    Args:
        coco_dset (kwcoco.CocoDataset): A KWCOCO dataset object.
        image_ids (List(int)): A list of image ids that belong in coco_dset.
        filtered_seasons (str | List(str) | None): Which seasons to not include in the
             returned image ids.
        ignore_winter_torrid_zone (bool, optional): Do not filter images within the
            Torrid region when winter is one of the filtered seasons. Defaults to True.

    Raises:
        ValueError: Check if the filtered seasons varible is a correctable type.
        ValueError: Check if one of the filtered seasons is not a valid season.

    Returns:
        List[int]: A list of filtered image ids. Should never be longer than the input
             image ids.

    Example:
        >>> # 0: Baseline run.
        >>> # xdoctest: +REQUIRES(env:DEVEL_TEST)
        >>> from geowatch.cli.coco_time_combine import *  # NOQA
        >>> import geowatch
        >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> # Load KWCOCO dataset.
        >>> input_kwcoco_fpath = data_dvc_dpath / 'Drop6/imgonly-AE_C001.kwcoco.json'
        >>> coco_dset = kwcoco.CocoDataset(input_kwcoco_fpath)
        >>> image_ids = coco_dset.images().gids
        >>> ignore_torrid_regions_gids = filter_image_ids_by_season(coco_dset,
        >>>                                image_ids,
        >>>                                filtered_seasons='winter',
        >>>                                ignore_winter_torrid_zone=True)
        >>> all_filtered_gids = filter_image_ids_by_season(coco_dset,
        >>>                       image_ids,
        >>>                       filtered_seasons='winter',
        >>>                       ignore_winter_torrid_zone=True)
        >>> assert len(all_filtered_gids) > len(ignore_torrid_regions_gids)
    """
    from kwutil import util_time
    # todo: find a way to abstract the domain specific concept of "season"
    hemipshere_to_season_map = {
        'northern': {
            'spring': [3, 4, 5],
            'summer': [6, 7, 8],
            'fall': [9, 10, 11],
            'winter': [12, 1, 2]
        },
        'southern': {
            'spring': [9, 10, 11],
            'summer': [12, 1, 2],
            'fall': [3, 4, 5],
            'winter': [6, 7, 8]
        }
    }

    # Check if there are any images to filter.
    if len(image_ids) == 0:
        print('WARNING: No images to filter.')
        return []

    # Check type of filtered_seasons variable and try to convert to list.
    if isinstance(filtered_seasons, str):
        filtered_seasons = [filtered_seasons]
    elif filtered_seasons is None:
        filtered_seasons = []
    elif isinstance(filtered_seasons, list):
        # Remove nones to workaround cli issue
        filtered_seasons = [f for f in filtered_seasons if f is not None]
        pass
    else:
        raise ValueError(f'Filtered seasons must be a list or string. Got "{type(filtered_seasons)}" type.')

    # Get seasons to filter (invalid seasons).
    filtered_seasons = set(filtered_seasons)
    valid_seasons = {'spring', 'summer', 'fall', 'winter'}

    invalid_seasons = filtered_seasons - valid_seasons
    if invalid_seasons:
        raise ValueError(f'Invalid seasons: {invalid_seasons}')

    # Get hemisphere of region.
    coco_img = coco_dset.coco_image(image_ids[0])
    import numpy as np
    geo_corner_coords = np.asarray(coco_img.img['geos_corners']['coordinates'])
    center_coord = geo_corner_coords.mean(axis=1)[0]
    _, lat = center_coord[0], center_coord[1]

    # Check if the latitude is in the northern or southern hemisphere.
    if lat > 0:
        hemisphere = 'northern'
    else:
        hemisphere = 'southern'

    # Check if the region is within the torrid zone.
    # Torrid zone is generally warm and less likely to contain snow.
    # https://en.wikipedia.org/wiki/Geographical_zone
    # https://www.toppr.com/guides/chemistry/environmental-chemistry/torrid-zone/
    # Should technically be 23.5 but bumpping up to 27 to include AE regions
    within_torrid_zone = abs(lat) < 27

    month_to_season_map = {}
    for season, months in hemipshere_to_season_map[hemisphere].items():
        for month in months:
            month_to_season_map[month] = season

    final_image_ids = []
    for image_id in image_ids:
        coco_img = coco_dset.coco_image(image_id)

        # Get month image was taken in.
        dt = util_time.coerce_datetime(coco_img['date_captured'])
        month = dt.month

        # Get season of month (depends on northern or southern hemisphere).
        img_season = month_to_season_map[month]

        # Do not filter image if ALL conditions are met:
        # 1) We are not removing images within the torrid zone during winter,
        # 2) the region latitude is within the torrid zone (|lat| < 23.5),
        # 3) winter is in the filtered seasons,
        # 4) the image was captured during winter.
        if ignore_winter_torrid_zone and within_torrid_zone and (img_season == 'winter') and ("winter" in filtered_seasons):
            final_image_ids.append(image_id)
        else:
            # Check if the season is not in the filtered seasons.
            if img_season not in filtered_seasons:
                final_image_ids.append(image_id)

    return final_image_ids


__cli__ = TimeCombineConfig

if __name__ == '__main__':
    """

    CommandLine:
        python -m geowatch.cli.coco_time_combine
    """
    main()



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub
import kwutil
import json
import rich
from rich.markup import escape


class Stage1EvaluateCLI(scfg.DataConfig):
    pred_fpath = scfg.Value(None, help='path to predicted file')
    true_fpath = scfg.Value(None, help='path to truth file')
    out_fpath = scfg.Value(None, help='path to evaluation file')
    workers = scfg.Value(0, help='number of parallel workers')

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
        rich.print('config = ' + escape(ub.urepr(config, nl=1)))

        data = {
            'info': [],
            'result': None
        }

        proc_context = kwutil.ProcessContext(
            name='stage1_evaluate',
            type='process',
            config=kwutil.Json.ensure_serializable(dict(config)),
            track_emissions=True,
        )
        proc_context.start()

        print('Load file')
        true_text = ub.Path(config.true_fpath).read_text()
        pred_text = ub.Path(config.pred_fpath).read_text()

        true_hashid = ub.hash_data(true_text)
        pred_hashid = ub.hash_data(pred_text)
        true_int = int(true_hashid, 16)
        pred_int = int(pred_hashid, 16)
        hamming_distance = bin(true_int ^ pred_int).count('1')
        size = (len(true_hashid) * 4)
        acc = (size - hamming_distance) / size

        metrics = {
            'accuracy': acc,
            'hamming_distance': hamming_distance,
        }

        # A dummy evaluate computation
        data['result'] = metrics

        obj = proc_context.stop()
        data['info'].append(obj)

        out_fpath = ub.Path(config.out_fpath)
        out_fpath.parent.ensuredir()
        out_fpath.write_text(json.dumps(data))
        print(f'wrote to: out_fpath={out_fpath}')

__cli__ = Stage1EvaluateCLI

if __name__ == '__main__':
    __cli__.main()


    r"""

    CommandLine:
        python ~/code/geowatch/docs/source/manual/tutorial/examples/mlops/mlops_example_module/cli/stage1_evaluate.py \
            --true_fpath ~/.bashrc \
            --pred_fpath ~/.bashrc \
            --out_fpath out.json

        python -m mlops_example_module.cli.stage1_predict
    """



import itertools
import ubelt as ub
import warnings
from functools import lru_cache
from typing import Iterable, Optional, Dict

try:
    from line_profiler import profile
except Exception:
    profile = ub.identity


def trackid_is_default(trackid):
    """
    Hack to decide if a trackid is really a site_id or if it was randomly
    assigned
    """
    if trackid is None:
        return True
    try:
        int(trackid)
        return True
    except ValueError:
        return False


def check_only_bg(category_sequence, bg_name=['No Activity']):
    if len(set(category_sequence) - set(bg_name)) == 0:
        return True
    else:
        return False


# --- geopandas utils ---
# make this a class?


def gpd_sort_by_gid(gdf, sorted_gids):

    dct = dict(zip(sorted_gids, range(len(sorted_gids))))
    gdf['gid_order'] = gdf['gid'].map(dct)

    # assert gdf['gid'].map(dct).groupby(
    # lambda x: x).is_monotonic_increasing.all()

    if gdf.empty:
        return gdf
    else:
        return gdf.groupby('track_idx', group_keys=False).apply(
            lambda x: x.sort_values('gid_order')).reset_index(drop=True).drop(
                columns=['gid_order'])


def gpd_len(gdf):
    return gdf['track_idx'].nunique()


@profile
def gpd_compute_scores(gdf, sub_dset, thrs: Iterable, ks: Dict, USE_DASK=False,
                       resolution=None, modulate=None):
    """
    TODO: This needs docs and examples for the BAS and SC/AC cases.

    Args:
        gdf (gdf.GeoDataFrame):
            input data frame tracks dataframe containing
            track_idx, gid, and poly columns.

        sub_dset (kwcoco.CocoDataset):
            dataset with reference to images

        thrs (List[float]):
            thresholds (-1) means take the average response, other values is
            the fraction of pixels with responses above that value.

        ks (Dict[str, List[str]]):
            mapping from "fg" to a list of "foreground classes"
            optionally also
            mapping from "bg" to a list of "background classes"

        resolution (str | None):
            resolution spec to compute scores at (e.g. "2GSD").

    Calls :func:`_compute_group_scores` on each dataframe row, which will
    execute the read for the image prediction scores for polygons with
    :func:`score_poly`.

    Returns:
        gdf.GeoDataFrame - a scored variant of the input data frame

             returns the per-channels scores as well as summed groups of
             channels. (not sure if that last one is necessary, might need to
             refactor to simplify)

    Example:
        >>> import kwcoco
        >>> from geowatch.tasks.tracking.utils import *  # NOQA
        >>> from geowatch.tasks.tracking.utils import _compute_group_scores, _build_annot_gdf
        >>> sub_dset = kwcoco.CocoDataset.demo('vidshapes1')
        >>> gdf, _ = _build_annot_gdf(sub_dset)
        >>> thrs = [-1, 'median']
        >>> ks = {'r|g': ['r', 'g'], 'bg': ['b']}
        >>> USE_DASK = 0
        >>> resolution = None
        >>> gdf2 = gpd_compute_scores(gdf, sub_dset, thrs, ks, USE_DASK, resolution)
    """
    import pandas as pd

    ks = {k: v for k, v in ks.items() if v}
    _valid_keys = list(set().union(itertools.chain.from_iterable(
        ks.values())))  # | ks.keys()

    # score_cols = list(itertools.product(_valid_keys, thrs))
    score_cols = [t[::-1] for t in itertools.product(thrs, _valid_keys)]

    if USE_DASK:  # 63% runtime
        import dask_geopandas
        # https://github.com/geopandas/dask-geopandas
        # _col_order = gdf.columns  # doesn't matter
        gdf = gdf.set_index('gid')
        # npartitions and chunksize are mutually exclusive
        gdf = dask_geopandas.from_geopandas(gdf, npartitions=8)
        meta = gdf._meta.join(pd.DataFrame(columns=score_cols, dtype=float))
        groups = gdf.groupby('gid', group_keys=False)
        gdf = groups.apply(_compute_group_scores,
                           thrs=thrs,
                           keys=_valid_keys,
                           meta=meta,
                           resolution=resolution,
                           sub_dset=sub_dset, modulate=modulate)
        # raises this, which is probably fine:
        # /home/local/KHQ/matthew.bernstein/.local/conda/envs/watch/lib/python3.9/site-packages/rasterio/features.py:362:
        # NotGeoreferencedWarning: Dataset has no geotransform, gcps, or rpcs.
        # The identity matrix will be returned.
        # _rasterize(valid_shapes, out, transform, all_touched, merge_alg)

        gdf = gdf.compute()  # gdf is now a GeoDataFrame again
        gdf = gdf.reset_index()
        # gdf = gdf.reindex(columns=_col_order)

    else:  # 95% runtime
        grouped = gdf.groupby('gid', group_keys=False)
        """
        grp = gdf.iloc[0:1]
        grp.name = grp.iloc[0].gid
        """
        gdf = grouped.apply(_compute_group_scores, thrs=thrs, _valid_keys=_valid_keys,
                            resolution=resolution, sub_dset=sub_dset, modulate=modulate)

    # fill nan scores from nodata pxls
    # groupby track instead of gid
    def _fillna(grp):
        if len(grp) == 0:
            grp = grp.reindex(list(ub.oset(grp.columns) | score_cols), axis=1)
        else:
            grp[score_cols] = grp[score_cols].fillna(method='ffill').fillna(0)
        return grp

    grouped = gdf.groupby('track_idx', group_keys=False)
    scored_gdf = grouped.apply(_fillna)

    # copy over to summed fg/bg channels
    for thr in thrs:
        for k, kk in ks.items():
            if kk:
                # https://github.com/pandas-dev/pandas/issues/20824#issuecomment-384432277
                sum_cols = [(ki, thr) for ki in kk]
                sum_cols = list(ub.oset(scored_gdf.columns) & sum_cols)
                scored_gdf[(k, thr)] = scored_gdf[sum_cols].sum(axis=1)

    return scored_gdf


def _compute_group_scores(grp, thrs=[], _valid_keys=[], resolution=None,
                          sub_dset=None, modulate=None):
    """
    Helper for :func:`gpd_compute_scores`.
    """
    import kwcoco
    import pandas as pd
    """
    Note:
        "name" is an attribute groupby only seems to give in the apply step.

        We can get a reference to the group object via:

            obj = list(grouped._iterate_slices())[0]
            ??? not sure if this is right

        The following is a MWE:

    Ignore:
        # Test groupby
        import pandas as pd
        df = pd.DataFrame({'a': [1, 2, 3], 'b': [2, 2, 1]})
        def foo(grp):
            print(f'grp.name={grp.name}')
            print(type(grp))
            print(f'grp={grp}')
        groups = df.groupby('b',group_keys=False)
        grp = list(groups._iterate_slices())[0]
        groups.apply(foo)

    Args:
        grp : A Pandas Group Object for data in the same iamge

    Example:
        >>> import kwcoco
        >>> from geowatch.tasks.tracking.utils import *  # NOQA
        >>> from geowatch.tasks.tracking.utils import _compute_group_scores, _build_annot_gdf
        >>> _valid_keys = ['r', 'g', 'b']
        >>> sub_dset = kwcoco.CocoDataset.demo('vidshapes1')
        >>> aids = list(sub_dset.images().annots[0])
        >>> grp, _ = _build_annot_gdf(sub_dset, aids=aids)
        >>> thrs = [-1, 'mean', 'max', 'min', 'median']
        >>> modulate = {'r': 0.0001}
        >>> gdf2 = _compute_group_scores(grp, thrs=thrs, _valid_keys=_valid_keys, sub_dset=sub_dset, modulate=modulate)
        >>> print(gdf2)
    """
    import numpy as np
    gid = getattr(grp, 'name', None)
    if gid is None:
        if len(grp) > 0:
            gid = grp.iloc[0]['gid']

    if gid is None:
        for thr in thrs:
            grp[[(k, thr) for k in _valid_keys]] = 0
    else:
        img = sub_dset.coco_image(gid)

        # Load the channels to score
        channels = kwcoco.FusedChannelSpec.coerce(_valid_keys)
        heatmaps_hwc = img.imdelay(channels, space='video', resolution=resolution).finalize()
        heatmaps_chw = heatmaps_hwc.transpose(2, 0, 1)
        if heatmaps_chw.dtype.kind != 'f':
            heatmaps_chw = heatmaps_chw.astype(np.float32)
        heatmaps_chw = np.ascontiguousarray(heatmaps_chw)
        if modulate is not None:
            chan_list = channels.to_list()
            for k, v in modulate.items():
                idx = chan_list.index(k)
                assert idx >= 0
                heatmaps_chw[idx] *= v

        assert isinstance(thrs, list)
        # score_cols = list(itertools.product(_valid_keys, thrs))
        score_cols = [t[::-1] for t in itertools.product(thrs, _valid_keys)]

        # Compute scores for each polygon.
        new_scores_rows = []
        for poly in grp['poly']:
            poly_scores_ = score_poly(poly, heatmaps_chw, threshold=thrs)
            # awk, making this serializable for kwcoco dataset
            poly_scores = list(ub.flatten(poly_scores_))
            col_to_score = dict(zip(score_cols, poly_scores))
            new_scores_rows.append(pd.Series(col_to_score))
        grp[score_cols] = new_scores_rows

        # scores = grp['poly'].apply(
        #     lambda p: pd.Series(dict(zip(
        #         score_cols,
        #         # awk, making this serializable for kwcoco dataset
        #         list(ub.flatten(score_poly(p, heatmaps_chw, threshold=thrs)))))
        #     ))
        # grp[score_cols] = scores
    return grp


# -----------------------


@profile
def score_track_polys(coco_dset,
                      video_id,
                      cnames=None,
                      score_chan=None,
                      resolution: Optional[str] = None):
    """
    Score the polygons in a kwcoco dataset based on heatmaps without modifying
    the polygon boundaries.

    Args:
        coco_dset (kwcoco.CocoDataset):

        video_id (int): video to score tracks for

        cnames (Iterable[str] | None):
            category names. Only annotations with these names will be
            considered.

        score_chan (kwcoco.ChannelSpec | None):
            score the track polygons by image overlap with this channel

    Note:
        This function needs a rename because we don't want this to mutate the
        kwcoco dataset ever.

    Returns:
        gpd.GeoDataFrame:
            With columns:
                gid: the image id
                poly: the polygon in video space
                track_idx: the annotation trackid

                And then for each score chan: c you get a column:
                    (c, -1) where the -1 indicates that no threshold was
                    applied, and that this is the mean of that channel
                    intensity under the polygon.

                Then there is another column where all channels are fused: f
                and you get a column: (f, -1)

    Note:
        The returned unerlying GDF should return polygons in video space as it
        will be consumed by :func:`_add_tracks_to_dset`.

    Example:
        >>> import kwcoco
        >>> coco_dset = kwcoco.CocoDataset.demo('vidshapes8-msi')
        >>> video_id = list(coco_dset.videos())[0]
        >>> cnames = None
        >>> resolution = None
        >>> score_chan = kwcoco.ChannelSpec.coerce('B1|B8|B8a|B10|B11')
        >>> gdf = score_track_polys(coco_dset, video_id, cnames, score_chan, resolution)
        >>> print(gdf)
    """
    # TODO could refactor to work on coco_dset.annots() and integrate
    import numpy as np

    aids = list(ub.flatten(coco_dset.images(video_id=video_id).annots))
    annots = coco_dset.annots(aids)
    # annots = coco_dset.annots()
    gdf, flat_scales = _build_annot_gdf(
        coco_dset, aids=aids, cnames=cnames, resolution=resolution)

    if score_chan is not None:
        # USE_DASK = True
        USE_DASK = False
        keys = {score_chan.spec: list(score_chan.unique())}
        sub_dset = coco_dset
        thrs = [-1]
        ks = keys
        gdf = gpd_compute_scores(gdf, sub_dset, thrs, ks, USE_DASK=USE_DASK,
                                 resolution=resolution)
    # TODO standard way to access sorted_gids
    sorted_gids = coco_dset.index._set_sorted_by_frame_index(
        np.unique(annots.gids))
    gdf = gpd_sort_by_gid(gdf, sorted_gids)

    if resolution is not None:
        # It should be the case that all of the scale factors are the same
        # because it is wrt to video space. Check for this and then
        # just apply a single warp.
        assert ub.allsame(flat_scales)
        if flat_scales:
            inverse_scale = 1 / flat_scales[0][0], 1 / flat_scales[0][1]
            gdf['poly'] = gdf['poly'].scale(
                xfact=inverse_scale[0],
                yfact=inverse_scale[1],
                origin=(0, 0, 0))

    return gdf


def _build_annot_gdf(coco_dset, aids=None, cnames=None, resolution=None):
    import geopandas as gpd
    import numpy as np

    annots = coco_dset.annots(aids)

    if cnames is not None:
        cnames = list(set(cnames))
        annots = annots.compress(
            np.in1d(np.array(annots.cnames, dtype=str), cnames))
        if len(annots) < 1:
            print(f'warning: no cnames={cnames} annots in dset dset.tag={coco_dset.tag}!')

    # Load polygon annotation segmentation in video space at the target
    # resolution
    gids = annots.images.ids
    gid_to_anns = ub.group_items(annots.objs, gids)

    flat_polys = []
    flat_gids = []
    flat_track_ids = []
    flat_scales = []
    for image_id, anns in gid_to_anns.items():
        coco_img = coco_dset.coco_image(image_id)

        img_polys = coco_img._annot_segmentations(anns, space='video',
                                                  resolution=resolution)
        flat_polys.extend(img_polys)
        flat_gids.extend([image_id] * len(img_polys))
        flat_track_ids.extend([ann['track_id'] for ann in anns])
        if resolution is not None:
            # Need to remember the inverse scale factor to get back to video
            # space.
            scale = coco_img._scalefactor_for_resolution(space='video',
                                                         resolution=resolution)
            flat_scales.append(scale)

    assert len(flat_polys) == len(annots), ('TODO handle multipolygon boundaries')

    flat_polys = [p.to_shapely() for p in flat_polys]
    gdf = gpd.GeoDataFrame({
        'gid': flat_gids,
        'poly': flat_polys,
        'track_idx': flat_track_ids,
    }, geometry='poly')
    return gdf, flat_scales


@lru_cache(maxsize=512)
def _rasterized_poly(shp_poly, h, w, pixels_are):
    import kwimage
    poly = kwimage.MultiPolygon.from_shapely(shp_poly)
    mask = poly.to_mask((h, w), pixels_are=pixels_are).data
    return mask


@profile
def score_poly(poly, probs, threshold=-1, use_rasterio=True):
    """
    Compute the average heatmap response of a heatmap inside of a polygon.

    Args:
        poly (kwimage.Polygon | MultiPolygon):
            in pixel coords

        probs (ndarray):
            heatmap to compare poly against in [..., c, h, w] format.
            The last two dimensions should be height, and width.
            Any leading batch dimensions will be preserved in output,
            e.g. (gid, chan, h, w) -> (gid, chan)

        use_rasterio (bool):
            use rasterio.features module instead of kwimage

        threshold (float | List[float | str]):
            Return fraction of poly with probs > threshold.  If -1, return
            average value of probs in poly. Can be a list of values, in which
            case returns all of them.

    Returns:
        List[ndarray] | ndarray:

            When thresholds is a list, returns a corresponding list of ndarrays
            with an entry keeping the leading dimensions of probs and
            marginalizing over the last two.

    Example:
        >>> import numpy as np
        >>> import kwimage
        >>> from geowatch.tasks.tracking.utils import score_poly
        >>> h = w = 64
        >>> poly = kwimage.Polygon.random().scale((w, h))
        >>> probs = np.random.rand(1, 3, h, w)
        >>> # Test with one threshold
        >>> threshold = [0.1, 0.2]
        >>> result = score_poly(poly, probs, threshold=threshold, use_rasterio=True)
        >>> print('result = {}'.format(ub.urepr(result, nl=1)))
        >>> # Test with multiple thresholds
        >>> threshold = 0.1
        >>> result = score_poly(poly, probs, threshold=threshold, use_rasterio=True)
        >>> print('result = {}'.format(ub.urepr(result, nl=1)))
        >>> # Test with -1 threshold
        >>> threshold = [-1, 'min', 'median']
        >>> result = score_poly(poly, probs, threshold=threshold, use_rasterio=True)
        >>> print('result = {}'.format(ub.urepr(result, nl=1)))

    Example:
        ### Grid of cases

        basis = {
            'threshold':
        }

    """
    import kwimage
    import numpy as np
    if not isinstance(poly, (kwimage.Polygon, kwimage.MultiPolygon)):
        poly = kwimage.MultiPolygon.from_shapely(poly)  # 2.4% of runtime

    _return_list = isinstance(threshold, Iterable)
    if not _return_list:
        threshold = [threshold]

    # First compute the valid bounds of the polygon
    # And create a mask for only the valid region of the polygon

    box = poly.box().quantize().to_xywh()

    # Ensure box is inside probs
    ymax, xmax = probs.shape[-2:]
    box = box.clip(0, 0, xmax, ymax).to_xywh()
    if box.area == 0:
        warnings.warn(
            'warning: scoring a polygon against an img with no overlap!')
        zeros = np.zeros(probs.shape[:-2])
        return [zeros] * len(threshold) if _return_list else zeros
    # sl_y, sl_x = box.to_slice()
    x, y, w, h = box.data
    pixels_are = 'areas' if use_rasterio else 'points'
    # kwimage inverse
    # 95% of runtime... would batch be faster?
    rel_poly = poly.translate((-x, -y))
    # rel_mask = rel_poly.to_mask((h, w), pixels_are=pixels_are).data
    # shapely polys hash correctly (based on shape, not memory location)
    # kwimage polys don't
    rel_mask = _rasterized_poly(rel_poly.to_shapely(), h, w, pixels_are)
    # Slice out the corresponding region of probabilities
    rel_probs = probs[..., y:y + h, x:x + w]

    result = []

    # handle nans
    msk = (np.isfinite(rel_probs) * rel_mask).astype(bool)
    all_non_finite = not msk.any()

    mskd_rel_probs = np.ma.array(rel_probs, mask=~msk)

    for t in threshold:
        if all_non_finite:
            stat = np.full(rel_probs.shape[:-2], fill_value=np.nan)
        elif t == 'max':
            stat = mskd_rel_probs.max(axis=(-2, -1)).filled(0)
        elif t == 'min':
            stat = mskd_rel_probs.min(axis=(-2, -1)).filled(0)
        elif t == 'mean':
            stat = mskd_rel_probs.mean(axis=(-2, -1)).filled(0)
        elif t == 'median':
            stat = np.ma.median(mskd_rel_probs, axis=(-2, -1)).filled(0)
        elif t == -1:
            # Alias for mean, todo: deprecate
            stat = mskd_rel_probs.mean(axis=(-2, -1)).filled(0)
        else:
            # Real threshold case
            hard_prob = rel_probs > t
            mskd_hard_prob = np.ma.array(hard_prob, mask=~msk)
            stat = mskd_hard_prob.mean(axis=(-2, -1)).filled(0)
        result.append(stat)

    return result if _return_list else result[0]


def _relative_mask_image(poly, image):
    """
    Unused, looking for a better API that can help extract image information
    under a polygon.

    Ignore:
        >>> import kwimage
        >>> poly = Polygon.random(rng=8).scale((128, 128)).translate(100, 100)
        >>> image = np.random.rand(512, 512, 3)
        >>> _relative_mask_image(poly, image)
    """
    import kwimage
    import numpy as np
    rel_mask, offset = poly.to_relative_mask(return_offset=True)
    box = kwimage.Box.coerce(list(offset) + list(rel_mask.shape[::-1]), format='xywh')
    rel_image = image[box.to_slice()]
    # Probably just want to return the hard mask, and the relative image?
    # Maybe we also return weights based on a distance transform?
    mask_1chan = rel_mask.data.astype(bool)
    mask_nchan = np.ones_like(rel_image, dtype=bool) * mask_1chan[:, :, None]
    mskd_rel_probs = np.ma.array(rel_image, mask=~mask_nchan)
    return mskd_rel_probs


@profile
def mask_to_polygons(probs,
                     thresh,
                     bounds=None,
                     use_rasterio=True,
                     thresh_hysteresis=None):
    """
    Extract a polygon from a 2D heatmap. Optionally within the bounds of
    another mask or polygon.

    Args:
        probs (ndarray): aka heatmap, image of probability values

        thresh (float): to turn probs into a hard mask

        bounds (kwimage.Polygon): a kwimage or shapely polygon to crop the results to

        use_rasterio (bool): use rasterio.features module instead of kwimage

        thresh_hysteresis: if not None, only keep polygons with at least one
            pixel of score >= thresh_hysteresis

    Yields:
        kwcoco.Polygon: extracted polygons.

    Example:
        >>> from geowatch.tasks.tracking.utils import mask_to_polygons
        >>> import kwimage
        >>> probs = kwimage.Heatmap.random(dims=(64, 64),
        >>>                                rng=0).data['class_probs'][0]
        >>> thresh = 0.5
        >>> polys = mask_to_polygons(probs, thresh)
        >>> poly1 = list(polys)[0]
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(probs > 0.5)

    Example:
        >>> from geowatch.tasks.tracking.utils import mask_to_polygons
        >>> import kwimage
        >>> import kwarray
        >>> rng = kwarray.ensure_rng(1043462368)
        >>> probs = kwimage.Heatmap.random(dims=(256, 256), rng=rng,
        >>>                                 ).data['class_probs'][0]
        >>> thresh = 0.5
        >>> polys1 = list(mask_to_polygons(
        >>>             probs, thresh, use_rasterio=0))
        >>> polys2 = list(mask_to_polygons(
        >>>             probs, thresh, use_rasterio=1))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> plt = kwplot.autoplt()
        >>> pnum_ = kwplot.PlotNums(nSubplots=4)
        >>> kwplot.imshow(probs, pnum=pnum_(), title='pixels_are=points')
        >>> for poly in polys1:
        >>>     poly.draw(facecolor='none', edgecolor='kitware_blue', alpha=0.5, linewidth=8)
        >>> kwplot.imshow(probs, pnum=pnum_(), title='pixels_are=areas')
        >>> for poly in polys2:
        >>>     poly.draw(facecolor='none', edgecolor='kitware_green', alpha=0.5, linewidth=8)
    """
    import kwimage
    import numpy as np
    import shapely.geometry
    from scipy.ndimage import label as ndm_label
    # Threshold scores
    if thresh_hysteresis is None:
        binary_mask = (probs > thresh).astype(np.uint8)
    else:
        mask = probs > thresh
        seeds = probs > thresh_hysteresis
        label_img = ndm_label(mask)[0]
        selected = np.unique(np.extract(seeds, label_img))
        binary_mask = np.isin(label_img, selected).astype(np.uint8)

    pixels_are = 'areas' if use_rasterio else 'points'
    if bounds is not None:
        try:
            bounds = shapely.geometry.shape(bounds)
        except ValueError:
            ...
        bounds_poly = kwimage.Polygon.coerce(bounds)
        bounds_mask = bounds_poly.to_mask(probs.shape, pixels_are=pixels_are)
        bounds_mask_ = bounds_mask.numpy().data.astype(np.uint8)
        binary_mask *= bounds_mask_

    final_mask = kwimage.Mask(binary_mask, 'c_mask')
    polygons = final_mask.to_multi_polygon(pixels_are=pixels_are)

    yield from polygons


def _validate_keys(key, bg_key):
    # for backwards compatibility
    if bg_key is None:
        bg_key = []
    key = list(key) if ub.iterable(key) else [key]
    bg_key = list(bg_key) if ub.iterable(bg_key) else [bg_key]

    # error checking
    if len(key) < 1:
        raise ValueError('must have at least one key')
    if (len(key) > len(set(key)) or len(bg_key) > len(set(bg_key))):
        raise ValueError('keys are duplicated')

    if 0:
        # Hack this off
        if not set(key).isdisjoint(set(bg_key)):
            raise ValueError('cannot have a key in foreground and background')

    return key, bg_key



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility script to perform net surgery on a model.

This script will perform net surgery on DeepLab models trained on a source
dataset and create a new checkpoint for the target dataset.
"""

from typing import Any, Dict, Text, Tuple

import numpy as np
import tensorflow as tf
from absl import app
from absl import flags
from absl import logging
from deeplab2 import common
from deeplab2 import config_pb2
from deeplab2.data import dataset
from deeplab2.model import deeplab
from google.protobuf import text_format

FLAGS = flags.FLAGS

flags.DEFINE_string('source_dataset', 'cityscapes',
                    'Dataset name on which the model has been pretrained. '
                    'Supported datasets: `cityscapes`.')

flags.DEFINE_string('target_dataset', 'motchallenge_step',
                    'Dataset name for conversion. Supported datasets: '
                    '`motchallenge_step`.')

flags.DEFINE_string('input_config_path', None,
                    'Path to a config file that defines the DeepLab model and '
                    'the checkpoint path.')

flags.DEFINE_string('output_checkpoint_path', None,
                    'Output filename for the generated checkpoint file.')

_SUPPORTED_SOURCE_DATASETS = {'cityscapes'}
_SUPPORTED_TARGET_DATASETS = {'motchallenge_step'}

_CITYSCAPES_TO_MOTCHALLENGE_STEP = (
    1,  # sidewalk
    2,  # building
    8,  # vegetation
    10,  # sky
    11,  # pedestrian
    12,  # rider
    18,  # bicycle
)

_DATASET_TO_INFO = {
    'cityscapes': dataset.CITYSCAPES_PANOPTIC_INFORMATION,
    'motchallenge_step': dataset.MOTCHALLENGE_STEP_INFORMATION,
}
_INPUT_SIZE = (1025, 2049, 3)


def _load_model(
        config_path: Text,
        source_dataset: Text) -> Tuple[deeplab.DeepLab, config_pb2.ExperimentOptions]:
    """Load DeepLab model based on config and dataset."""
    options = config_pb2.ExperimentOptions()
    with tf.io.gfile.GFile(config_path) as f:
        text_format.Parse(f.read(), options)
    options.model_options.panoptic_deeplab.semantic_head.output_channels = (
        _DATASET_TO_INFO[source_dataset].num_classes)
    model = deeplab.DeepLab(options,
                            _DATASET_TO_INFO[source_dataset])
    return model, options


def _convert_bias(input_tensor: np.ndarray,
                  label_list: Tuple[int, ...]) -> np.ndarray:
    """Converts 1D tensor bias w.r.t. label list.

    We select the subsets from the input_tensor based on the label_list.

    We assume input_tensor has shape = [num_classes], where
    input_tensor is the bias weights trained on source dataset, and num_classes
    is the number of classes in source dataset.

    Args:
      input_tensor: A numpy array with ndim == 1.
      label_list: A tuple of labels used for net surgery.

    Returns:
      A numpy array with values modified.

    Raises:
      ValueError: input_tensor's ndim != 1.
    """
    if input_tensor.ndim != 1:
        raise ValueError('The bias tensor should have ndim == 1.')

    num_elements = len(label_list)
    output_tensor = np.zeros(num_elements, dtype=np.float32)
    for i, label in enumerate(label_list):
        output_tensor[i] = input_tensor[label]
    return output_tensor


def _convert_kernels(input_tensor: np.ndarray,
                     label_list: Tuple[int, ...]) -> np.ndarray:
    """Converts 4D tensor kernels w.r.t. label list.

    We select the subsets from the input_tensor based on the label_list.

    We assume input_tensor has shape = [h, w, input_dim, num_classes], where
    input_tensor is the kernel weights trained on source dataset, and num_classes
    is the number of classes in source dataset.

    Args:
      input_tensor: A numpy array with ndim == 4.
      label_list: A tuple of labels used for net surgery.

    Returns:
      A numpy array with values modified.

    Raises:
      ValueError: input_tensor's ndim != 4.
    """
    if input_tensor.ndim != 4:
        raise ValueError('The kernels tensor should have ndim == 4.')

    num_elements = len(label_list)
    kernel_height, kernel_width, input_dim, _ = input_tensor.shape
    output_tensor = np.zeros(
        (kernel_height, kernel_width, input_dim, num_elements), dtype=np.float32)
    for i, label in enumerate(label_list):
        output_tensor[:, :, :, i] = input_tensor[:, :, :, label]
    return output_tensor


def _restore_checkpoint(restore_dict: Dict[Any, Any],
                        options: config_pb2.ExperimentOptions
                        ) -> tf.train.Checkpoint:
    """Reads the provided dict items from the checkpoint specified in options.

    Args:
      restore_dict: A mapping of checkpoint item to location.
      options: A experiment configuration containing the checkpoint location.

    Returns:
      The loaded checkpoint.
    """
    ckpt = tf.train.Checkpoint(**restore_dict)
    if tf.io.gfile.isdir(options.model_options.initial_checkpoint):
        path = tf.train.latest_checkpoint(
            options.model_options.initial_checkpoint)
        status = ckpt.restore(path)
    else:
        status = ckpt.restore(options.model_options.initial_checkpoint)
    status.expect_partial().assert_existing_objects_matched()
    return ckpt


def main(_) -> None:
    if FLAGS.source_dataset not in _SUPPORTED_SOURCE_DATASETS:
        raise ValueError('Source dataset is not supported. Use --help to get list '
                         'of supported datasets.')
    if FLAGS.target_dataset not in _SUPPORTED_TARGET_DATASETS:
        raise ValueError('Target dataset is not supported. Use --help to get list '
                         'of supported datasets.')

    logging.info('Loading DeepLab model from config %s', FLAGS.input_config_path)
    source_model, options = _load_model(FLAGS.input_config_path,
                                        FLAGS.source_dataset)
    logging.info('Load pretrained checkpoint.')
    _restore_checkpoint(source_model.checkpoint_items, options)
    source_model(tf.keras.Input(_INPUT_SIZE), training=False)

    logging.info('Perform net surgery.')
    semantic_weights = (
        source_model._decoder._semantic_head.final_conv.get_weights())  # pylint: disable=protected-access

    if (FLAGS.source_dataset == 'cityscapes' and
            FLAGS.target_dataset == 'motchallenge_step'):
        # Kernels.
        semantic_weights[0] = _convert_kernels(semantic_weights[0],
                                               _CITYSCAPES_TO_MOTCHALLENGE_STEP)
        # Bias.
        semantic_weights[1] = _convert_bias(semantic_weights[1],
                                            _CITYSCAPES_TO_MOTCHALLENGE_STEP)

    logging.info('Load target model without last semantic layer.')
    target_model, _ = _load_model(FLAGS.input_config_path, FLAGS.target_dataset)
    restore_dict = target_model.checkpoint_items
    del restore_dict[common.CKPT_SEMANTIC_LAST_LAYER]

    ckpt = _restore_checkpoint(restore_dict, options)
    target_model(tf.keras.Input(_INPUT_SIZE), training=False)
    target_model._decoder._semantic_head.final_conv.set_weights(semantic_weights)  # pylint: disable=protected-access

    logging.info('Save checkpoint to output path: %s',
                 FLAGS.output_checkpoint_path)
    ckpt = tf.train.Checkpoint(**target_model.checkpoint_items)
    ckpt.save(FLAGS.output_checkpoint_path)


if __name__ == '__main__':
    flags.mark_flags_as_required(
        ['input_config_path', 'output_checkpoint_path'])
    app.run(main)



"""
Basline Example:

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
    MAE_MODEL_FPATH="$DVC_EXPT_DPATH/models/wu/wu_mae_2023_04_21/Drop6-epoch=01-val_loss=0.20.ckpt"
    KWCOCO_BUNDLE_DPATH=$DVC_DATA_DPATH/Drop7-MedianNoWinter10GSD

    python -m geowatch.utils.simple_dvc request "$MAE_MODEL_FPATH"

    # NOTE: different predict files correspond to different models
    # TODO: make the model size a parameter (or better yet inferred)

    python -m geowatch.tasks.mae.predictV3 \
        --device="cuda:0"\
        --mae_ckpt_path="$MAE_MODEL_FPATH"\
        --input_kwcoco="$KWCOCO_BUNDLE_DPATH/imganns-KR_R001.kwcoco.zip"\
        --output_kwcoco="$KWCOCO_BUNDLE_DPATH/imganns-KR_R001-testmae.kwcoco.zip"\
        --window_space_scale=1.0 \
        --workers=8 \
        --io_workers=8

    # After your model predicts the outputs, you should be able to use the
    # geowatch visualize tool to inspect your features.
    python -m geowatch visualize "$KWCOCO_BUNDLE_DPATH/imganns-KR_R001-testmae.kwcoco.zip" \
        --channels "red|green|blue,mae.8:11,mae.14:17" --stack=only --workers=avail --animate=True \
        --draw_anns=False

"""
import ubelt as ub
import scriptconfig as scfg
import albumentations as A
import kwcoco
import kwimage
import ndsampler
import sys
import torch
import torch.nn as nn
from pytorch_lightning import LightningModule
from torch.nn import L1Loss as MSE
from torch.utils.data import DataLoader, Dataset
from einops import rearrange
from einops.layers.torch import Rearrange
import numpy as np
from kwutil import util_parallel
from geowatch.tasks.fusion.predict import CocoStitchingManager


class WatchDataset(Dataset):
    S2_l2a_channel_names = [
        'B02.tif', 'B01.tif', 'B03.tif', 'B04.tif', 'B05.tif', 'B06.tif', 'B07.tif', 'B08.tif', 'B09.tif', 'B11.tif', 'B12.tif', 'B8A.tif'
    ]
    S2_channel_names = [
        'coastal', 'blue', 'green', 'red', 'B05', 'B06', 'B07', 'nir', 'B09', 'cirrus', 'swir16', 'swir22', 'B8A'
    ]
    L8_channel_names = [
        'coastal', 'lwir11', 'lwir12', 'blue', 'green', 'red', 'nir', 'swir16', 'swir22', 'pan', 'cirrus'
    ]

    def __init__(self, coco_dset, sensor=['S2'], bands=['shared'],
                 segmentation=False, patch_size=224, mask_patch_size=16, num_images=2,
                 mode='train', patch_overlap=.25, bas=True, rng=None, mask_pct=.5, mask_time_width=2,
                 temporal_mode='cat', window_space_scale=1.0):
        super().__init__()
        if not isinstance(bands, list):
            bands = [bands]
        if not isinstance(sensor, list):
            sensor = [sensor]
        assert (temporal_mode in ['cat', 'stack'])

        # initialize dataset
        print('load dataset')
        self.coco_dset: kwcoco.CocoDataset = kwcoco.CocoDataset.coerce(coco_dset)

        print('filter dataset')
        # Filter out worldview images (better to use subset than remove)
        images: kwcoco.coco_objects1d.Images = self.coco_dset.images()
        flags = [s in sensor for s in images.lookup('sensor_coarse')]
        valid_image_ids : list[int] = list(images.compress(flags))
        self.coco_dset = self.coco_dset.subset(valid_image_ids)

        self.images : kwcoco.coco_objects1d.Images = self.coco_dset.images()
        self.sampler = ndsampler.CocoSampler(self.coco_dset)

        window_dims = [patch_size, patch_size]
        time_dims = num_images

        NEW_GRID = 1
        if NEW_GRID:
            print('make grid')
            from geowatch.tasks.fusion.datamodules.kwcoco_video_data import sample_video_spacetime_targets
            sample_grid = sample_video_spacetime_targets(
                self.coco_dset, window_dims=window_dims,
                time_dims=time_dims, window_overlap=patch_overlap,
                time_sampling='hardish3', time_span='1y',
                use_annot_info=False,
                keepbound=True,
                use_centered_positives=False,
                window_space_scale=window_space_scale
            )

            samples = sample_grid['targets']
            for tr in samples:
                tr['vidid'] = tr['video_id']  # hack
            print('made grid')
        else:
            grid = self.sampler.new_sample_grid(**{
                'task': 'video_detection',
                'window_dims': [num_images, patch_size, patch_size],
                'window_overlap': patch_overlap,
            })
            if segmentation:
                samples = grid['positives']
            else:
                samples = grid['positives'] + grid['negatives']

        # vidid_to_patches = ub.group_items(samples, key=lambda x: x['vidid'])
        # self.vidid_to_patches = vidid_to_patches
        print('build patches')
        grouped = ub.group_items(
                samples,
                lambda x: tuple(
                    [x['vidid']] + [gid for gid in x['gids']]
                )
                )
        grouped = ub.sorted_keys(grouped)
        self.patches : list[dict] = list(ub.flatten(grouped.values()))
        self.bands = []
        # no channels selected
        if len(bands) < 1:
            raise ValueError(f'bands must be specified. Options are {", ".join(bands)}, or all')
        # all channels selected
        elif len(bands) == 1:
            if bands[0].lower() == 'all':
                self.bands = bands
            elif bands[0].lower() == 'shared':
                self.bands = ['red', 'green', 'blue', 'nir', 'swir16', 'swir22']
            elif bands[0] == 'r|g|b':
                self.bands.append('r|g|b')

        self.num_channels = len(self.bands)
        self.bands = "|".join(self.bands)

        # define augmentations
        print('build augs')
        additional_targets = dict()
        self.num_images = num_images

        for i in range(self.num_images):
            additional_targets['image{}'.format(1 + i)] = 'image'
            additional_targets['seg{}'.format(i + 1)] = 'mask'

        self.transforms = A.NoOp()

        self.mode = mode
        self.segmentation = segmentation
        self.patch_size = patch_size
        self.bas = bas
        if self.bas:
            self.positive_indices = [0, 1, 3]
            self.ignore_indices = [2, 6]
        else:
            self.positive_indices = [0, 1, 2, 3]
            self.ignore_indices = [6]
        print('finished dataset init')

        self.temporal_mode = temporal_mode

    def __len__(self):
        return len(self.patches)

    def __getitem__(self, idx):
        #if idx > 500: raise IndexError
        tr : dict = self.patches[idx]
        tr['channels'] = self.bands
        tr = self.update_target_properties(tr)
        # vidid = tr['vidid']
        gids : list[int] = tr['gids']

        sample = self.sampler.load_sample(tr, nodata='float')
        images : np.ndarray = sample['im']
        std = np.nanstd(images)
        mean = np.nanmean(images)
        if std != 0:
            images = np.nan_to_num((images - mean) / std)
        else:
            images = np.zeros_like(images)

        if self.temporal_mode == 'cat':
            images = torch.cat([torch.tensor(x) for x in images], dim=0).permute(2, 0, 1)
        else:
            images = torch.tensor(images).permute(0, 3, 1, 2)

        vidspace_box = kwimage.Box.from_slice(tr['space_slice'])

        scale_outspace_from_vidspace = tr['scale'] / 4  # Add it back
        outspace_box = vidspace_box.scale(scale_outspace_from_vidspace).quantize().astype(np.int32)
        item = dict()
        im1_id = gids[0]
        img_obj1 : dict = self.coco_dset.index.imgs[im1_id]
        video_obj = self.coco_dset.index.videos[img_obj1['video_id']]

        full_stitch_vidspace_box = kwimage.Box.coerce([0, 0, video_obj['width'], video_obj['height']], format='xywh')
        full_stitch_outspace_box = full_stitch_vidspace_box.scale(scale_outspace_from_vidspace).quantize().astype(np.int32)

        item['full_stitch_outspace_ltrb'] = torch.from_numpy(full_stitch_outspace_box.data)
        item['sample_outspace_ltrb'] = torch.from_numpy(outspace_box.data)
        item['scale_outspace_from_vid'] = scale_outspace_from_vidspace

        return images, item

    def update_target_properties(self, target):
        """
        Populate the target so it has the correct input scale and bands.
        """
        # Handle target scale
        from geowatch.tasks.fusion.datamodules import data_utils
        gids : list[int] = target['gids']
        im1_id = gids[0]
        img_obj1 : dict = self.coco_dset.index.imgs[im1_id]
        video_obj = self.coco_dset.index.videos[img_obj1['video_id']]
        vidspace_gsd = video_obj.get('target_gsd', None)
        resolved_input_scale = data_utils.resolve_scale_request(request=1.0, data_gsd=vidspace_gsd)
        target['scale'] = resolved_input_scale['scale']
        target['channels'] = self.bands
        target['_input_gsd'] = resolved_input_scale['gsd']
        target['_native_video_gsd'] = resolved_input_scale['data_gsd']
        return target


class MAEPredictConfig(scfg.DataConfig):
    """
    Configuration for WashU MAE models
    """
    device = scfg.Value('cuda:0', type=str)
    mae_ckpt_path = scfg.Value(None, type=str)
    batch_size = scfg.Value(1, type=int)
    workers = scfg.Value(4, help=ub.paragraph(
            '''
            number of background data loading workers
            '''), alias=['num_workers'])
    io_workers = scfg.Value(8, help=ub.paragraph(
            '''
            number of background data writing workers
            '''), alias=['write_workers'])

    window_resolution = scfg.Value(1.0, help='The window GSD to build the grid at', alias=['window_space_scale'])

    sensor = scfg.Value(['S2', 'L8'], nargs='+')
    bands = scfg.Value(['shared'], type=str, help=ub.paragraph(
            '''
            Choose bands on which to train. Can specify 'all' for all
            bands from given sensor, or 'share' to use common bands when
            using both S2 and L8 sensors
            '''), nargs='+')
    patch_overlap = scfg.Value(0.25, type=float)
    input_kwcoco = scfg.Value(None, type=str, required=True, help=ub.paragraph(
            '''
            Path to kwcoco dataset with images to generate feature for
            '''))
    output_kwcoco = scfg.Value(None, type=str, required=True, help=ub.paragraph(
            '''
            Path to write an output kwcoco file. Output file will be a
            copy of input_kwcoco with addition feature fields generated
            by predict.py rerooted to point to the original data.
            '''))

    assets_dname = scfg.Value('_assets', help=ub.paragraph(
        '''
        The name of the top-level directory to write new assets.
        '''))


def pair(t):
    return t if isinstance(t, tuple) else (t, t)


class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)


class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout=0.):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, dim),
            nn.Dropout(dropout)
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=64, dropout=0.):
        super().__init__()
        inner_dim = dim_head *  heads
        project_out = not (heads == 1 and dim_head == dim)

        self.heads = heads
        self.scale = dim_head ** -0.5

        self.attend = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)

        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim),
            nn.Dropout(dropout)
        ) if project_out else nn.Identity()

    def forward(self, x):
        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.heads), qkv)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn = self.attend(dots)
        attn = self.dropout(attn)

        out = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        return self.to_out(out)


class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, mlp_dim, dropout=0.):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                PreNorm(dim, Attention(dim, heads=heads, dim_head=dim_head, dropout=dropout)),
                PreNorm(dim, FeedForward(dim, mlp_dim, dropout=dropout))
            ]))

    def forward(self, x):
        for attn, ff in self.layers:
            x = attn(x) + x
            x = ff(x) + x
        return x


class ViT(nn.Module):
    def __init__(self, *, image_size, image_patch_size, frames, frame_patch_size, dim, depth, heads, mlp_dim, channels=6, dim_head=64, dropout=0., emb_dropout=0.):
        super().__init__()
        image_height, image_width = pair(image_size)
        patch_height, patch_width = pair(image_patch_size)

        assert image_height % patch_height == 0 and image_width % patch_width == 0, 'Image dimensions must be divisible by the patch size.'
        assert frames % frame_patch_size == 0, 'Frames must be divisible by frame patch size'

        num_patches = (image_height // patch_height) * (image_width // patch_width) * (frames // frame_patch_size)
        patch_dim = channels * patch_height * patch_width * frame_patch_size

        self.to_patch_embedding = nn.Sequential(
            Rearrange('b (f pf) c (h p1) (w p2) -> b (f h w) (p1 p2 pf c)', p1=patch_height, p2=patch_width, pf=frame_patch_size),
            nn.Linear(patch_dim, dim),
        )

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, dim))

        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = Transformer(dim, depth, heads, dim_head, mlp_dim, dropout)

    def forward(self, video):
        x = self.to_patch_embedding(video)
        b, n, _ = x.shape

        x += self.pos_embedding[:, :(n + 1)]
        x = self.dropout(x)

        x = self.transformer(x)

        return x


class MAE(nn.Module):
    def __init__(
        self,
        *,
        encoder,
        decoder_dim,
        masking_ratio=0.75,
        decoder_depth=8,
        decoder_heads=8,
        decoder_dim_head=64
    ):
        super().__init__()
        assert masking_ratio > 0 and masking_ratio < 1, 'masking ratio must be kept between 0 and 1'
        self.masking_ratio = masking_ratio

        # extract some hyperparameters and functions from encoder (vision transformer to be trained)

        self.encoder = encoder
        num_patches, encoder_dim = encoder.pos_embedding.shape[-2:]
        self.to_patch, self.patch_to_emb = encoder.to_patch_embedding[:2]
        pixel_values_per_patch = self.patch_to_emb.weight.shape[-1]

        # decoder parameters
        self.decoder_dim = decoder_dim
        self.enc_to_dec = nn.Linear(encoder_dim, decoder_dim) if encoder_dim != decoder_dim else nn.Identity()
        self.mask_token = nn.Parameter(torch.randn(decoder_dim))
        self.decoder = Transformer(dim=decoder_dim, depth=decoder_depth, heads=decoder_heads, dim_head=decoder_dim_head, mlp_dim=decoder_dim * 4)
        #self.decoder_pos_emb = nn.Embedding(num_patches, decoder_dim)
        self.decoder_pos_emb = nn.Parameter(torch.randn(num_patches, decoder_dim))
        self.to_pixels = nn.Linear(decoder_dim, pixel_values_per_patch)
        self.out = nn.Sigmoid()

    def forward(self, img):

        patches = self.to_patch(img)

        tokens = self.patch_to_emb(patches)
        tokens = tokens + self.encoder.pos_embedding[:, 1:(tokens.shape[1] + 1)]

        encoded_tokens = self.encoder.transformer(tokens)

        encoded_tokens = rearrange(encoded_tokens, 'b (f h w) d -> b f h w d', h=32, w=32)

        return encoded_tokens


class MaeCityscape(LightningModule):
    def __init__(self, dataset, **kwargs):
        super().__init__()
        self.vit = ViT(
            image_size=128,
            image_patch_size=4,
            frames=4,
            frame_patch_size=1,
            dim=64,
            depth=12,
            heads=12,
            mlp_dim=1024,
            dropout=0.1
        )
        self.model = MAE(
            encoder=self.vit,
            masking_ratio=0.90,
            decoder_dim=128,
            decoder_depth=6,
        )
        self.dataset = dataset
        self.batch_size = kwargs.get('batch_size', 4)
        self.num_workers = kwargs.get('num_workers', 16)
        self.lr = kwargs.get('lr', 0.02)
        self.acc = MSE()

    def forward(self, x):
        return self.model(x)

    def shared_step(self, batch, batch_idx):
        (x, masks), dates = batch

        pred, gt, viz, mi, ui = self(x)
        #gt = repeat(gt, 'b n d -> b (n n2) d', n2= 2)
        batch_range = torch.arange(x.shape[0], device=x.device)[:, None]
        loss = 0.999 * self.acc(pred[batch_range, mi], gt[batch_range, mi]) + 0.001 * self.acc(pred[batch_range, ui], gt[batch_range, ui])
        #loss = self.acc(pred, gt)
        return loss, viz, pred, gt


def sigmoid(a):
    return 1 / (1 + np.exp(-a))


class Predict():
    def __init__(self, args):

        self.device = args.device
        self.data_path = args.input_kwcoco
        self.dataset = WatchDataset(self.data_path, sensor=args.sensor, bands=args.bands,
                                    segmentation=False, patch_size=128, mask_patch_size=16, num_images=4,
                                    mode='train', mask_pct=0.5, patch_overlap=args.patch_overlap,
                                    temporal_mode='stack',
                                    mask_time_width=2, window_space_scale=args.window_resolution)

        print("Dataset load finished ...")

        self.model = MaeCityscape(self.dataset)

        self.model = self.model.load_from_checkpoint(args.mae_ckpt_path, dataset=self.dataset)
        print("Model load finished ...")

        self.dataloader = DataLoader(self.dataset,
                                     shuffle=False,
                                     batch_size=1,
                                     num_workers=args.workers,
                                     persistent_workers=False,
                                     pin_memory=True)

        print('copy dataset')
        self.output_dset = self.dataset.coco_dset.copy()

        print('reroot')
        self.output_dset.reroot(absolute=True)
        self.output_dset.fpath = args.output_kwcoco
        self.output_dset.reroot(absolute=False)

        self.save_channels = 'mae.0:16'
        self.output_kwcoco_path = ub.Path(self.output_dset.fpath)
        out_folder = self.output_kwcoco_path.parent
        self.output_feat_dpath = (out_folder / args.assets_dname).ensuredir()

        self.imwrite_kw = {
            'compress': 'DEFLATE',
            'backend': 'gdal',
            'blocksize': 128,
        }

        self.stitch_manager = CocoStitchingManager(
            result_dataset=self.output_dset,
            short_code=args.assets_dname,
            chan_code=self.save_channels,
            stiching_space='video',
            prob_compress=self.imwrite_kw['compress'],
            quantize=True,
        )

        from geowatch.utils import process_context
        self.proc_context = process_context.ProcessContext(
            args=sys.argv,
            type='process',
            name='geowatch.tasks.mae.predict',
        )

    def __call__(self):

        writer_queue = util_parallel.BlockingJobQueue(max_workers=4)
        self.stitch_manager.writer_queue = writer_queue

        self.proc_context.start()

        self.proc_context.add_disk_info(ub.Path(self.output_dset.fpath).parent)
        self.output_dset.dataset.setdefault('info', [])
        self.output_dset.dataset['info'].append(self.proc_context.obj)

        print('Evaluating and saving features')

        self.model.eval()
        self.model.to(self.device)
        num_batches = len(self.dataloader)
        preds = []
        with torch.no_grad():
            seen_images = set()
            prog = ub.ProgIter(enumerate(self.dataloader), total=num_batches, desc='Compute features', verbose=1)
            for batch_idx, batch in prog:
                x, item = batch
                x = x.to(self.device)
                #x2 = rearrange(x, 'b (f pf) c h w -> b (pf f) c h w', pf=2)

                pred = self.model(x)
                #pred2 = self.model(x2)
                preds = pred.cpu().detach().numpy()
                #preds2 = pred2.cpu().detach().numpy()

                target = self.dataset.patches[batch_idx]

                new_complete_gids = target.get('new_complete_gids', [])

                for gid in new_complete_gids:
                    assert gid not in seen_images
                    seen_images.add(gid)
                    self.stitch_manager.submit_finalize_image(gid)

                gid1, gid2, gid3, gid4 = target['gids']

                sample_outspace_ltrb = kwimage.Box.coerce(item['sample_outspace_ltrb'].numpy(), format='ltrb')
                full_stitch_outspace_box = kwimage.Box.coerce(item['full_stitch_outspace_ltrb'].numpy(), format='ltrb')
                scale_outspace_from_vid = item['scale_outspace_from_vid'].numpy()[0]
                outspace_slice = sample_outspace_ltrb.to_slice()
                outspace_dsize = full_stitch_outspace_box.dsize

                feat1 = preds[:, 0, :, :, :].squeeze()
                feat2 = preds[:, 1, :, :, :].squeeze()
                feat3 = preds[:, 2, :, :, :].squeeze()
                feat4 = preds[:, 3, :, :, :].squeeze()

                self.stitch_manager.accumulate_image(
                    gid1, outspace_slice, feat1,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

                self.stitch_manager.accumulate_image(
                    gid2, outspace_slice, feat2,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

                self.stitch_manager.accumulate_image(
                    gid3, outspace_slice, feat3,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

                self.stitch_manager.accumulate_image(
                    gid4, outspace_slice, feat4,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

            print('Finalize already compelted jobs')
            writer_queue.wait_until_finished(desc='Finalize submitted jobs')

            for gid in ub.ProgIter(list(self.stitch_manager.image_stitchers.keys()), desc='submit loose write jobs'):
                if gid not in seen_images:
                    seen_images.add(gid)
                    self.stitch_manager.submit_finalize_image(gid)

            print('Finalize loose jobs')
            writer_queue.wait_until_finished()

        print('Finish process context')
        self.proc_context.add_device_info(self.device)
        self.proc_context.stop()

        print('Write to dset.fpath = {!r}'.format(self.output_dset.fpath))
        self.output_dset.dump(self.output_dset.fpath, newlines=True)
        print('Done')

        return


def main():
    args = MAEPredictConfig.cli()
    predict = Predict(args)
    predict()


if __name__ == '__main__':
    """
    SeeAlso:
        ../../cli/queue_cli/prepare_teamfeats.py

        # Team Features on Drop3
        DVC_DPATH=$(geowatch_dvc)
        KWCOCO_BUNDLE_DPATH=$DVC_DPATH/Aligned-Drop3-TA1-2022-03-10
        python -m geowatch.cli.queue_cli.prepare_teamfeats \
            --base_fpath=$KWCOCO_BUNDLE_DPATH/data.kwcoco.json \
            --with_depth=0 \
            --with_landcover=0 \
            --with_materials=0  \
            --with_invariants=1 \
            --do_splits=0 \
            --gres=0 --backend=serial --run=1

    CommandLine:
        python -m geowatch.tasks.template.predict --help

        DVC_DPATH=$(geowatch_dvc)
        PRETEXT_PATH=$DVC_DPATH/models/uky/uky_invariants_2022_02_11/TA1_pretext_model/pretext_package.pt
        SSEG_PATH=$DVC_DPATH/models/uky/uky_invariants_2022_02_11/TA1_segmentation_model/segmentation_package.pt
        PCA_FPATH=$DVC_DPATH/models/uky/uky_invariants_2022_02_11/TA1_pretext_model/pca_projection_matrix.pt
        KWCOCO_BUNDLE_DPATH=$DVC_DPATH/Drop2-Aligned-TA1-2022-02-15
        python -m geowatch.tasks.invariants.predict \
            --pretext_package_path "$PRETEXT_PATH" \
            --segmentation_package_path "$SSEG_PATH" \
            --pca_projection_path "$PCA_FPATH" \
            --input_kwcoco $KWCOCO_BUNDLE_DPATH/data.kwcoco.json \
            --workers=avail \
            --do_pca 0 \
            --patch_overlap=0.3 \
            --output_kwcoco $KWCOCO_BUNDLE_DPATH/uky_invariants.kwcoco.json \
            --tasks before_after pretext

        python -m geowatch stats $KWCOCO_BUNDLE_DPATH/uky_invariants.kwcoco.json

        python -m geowatch visualize $KWCOCO_BUNDLE_DPATH/uky_invariants/invariants_nowv_vali.kwcoco.json \
            --channels "invariants.7,invariants.6,invariants.5" --animate=True \
            --select_images '.sensor_coarse != "WV"' --draw_anns=False

    Ignore:
        ### Command 1 / 2 - geowatch-teamfeat-job-0
        python -m geowatch.tasks.invariants.predict \
            --input_kwcoco "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/data_kr1br2.kwcoco.json" \
            --output_kwcoco "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/data_kr1br2_uky_invariants.kwcoco.json" \
            --pretext_package_path "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/uky/uky_invariants_2022_03_21/pretext_model/pretext_package.pt" \
            --pca_projection_path  "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/uky/uky_invariants_2022_03_21/pretext_model/pretext_pca_104.pt" \
            --do_pca 0 \
            --patch_overlap=0.0 \
            --workers="2" \
            --io_workers 0 \
            --tasks before_after pretext

        cd /home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC
        kwcoco subset --src=data.kwcoco.json --dst=AE_R001.kwcoco.json --select_videos='.name == "AE_R001"'
        kwcoco subset --src=data.kwcoco.json --dst=NZ_R001.kwcoco.json --select_videos='.name == "NZ_R001"'

        python -m geowatch.tasks.mae.predict \
        --device="cuda:0"\
        --mae_ckpt_path="/storage1/fs1/jacobsn/Active/user_s.sastry/smart_watch/new_models/checkpoints/Drop6-epoch=01-val_loss=0.20.ckpt"\
        --input_kwcoco="$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/data_train_I2L_split6.kwcoco.zip"\
        --output_kwcoco="$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/mae_v1_train_split6.kwcoco.zip"\
        --window_space_scale=1.0 \
        --workers=8 \
        --io_workers=8

        python -m geowatch visualize $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/mae_v1_train_split6.kwcoco.zip \
        --channels "red|green|blue,mae.8:11,mae.14:17" --stack=only --workers=avail --animate=True \
        --draw_anns=False


    """
    main()
